"""Portuguese translations of the program interface"""

MSG_WELCOME_1 = "Bem vindo ao Calcure"
MSG_WELCOME_2 = "Seu calendário e gerenciador de tarefas do terminal!"
MSG_WELCOME_3 = "Configuração e arquivos de dados criados em:"
MSG_WELCOME_4 = "Para apoio, contribuição, e informações adicionais, visitar:"
MSG_WELCOME_5 = "Aperte ? para ver todas as combinações de teclas ou qualquer outra tecla para continuar."

TITLE_KEYS_GENERAL = "COMBINAÇÃO DE TECLAS GERAIS"
TITLE_KEYS_CALENDAR = "COMBINAÇÃO DE TECLAS DO CALENDÁRIO"
TITLE_KEYS_JOURNAL  = "COMBINAÇÃO DE TECLAS DO DIÁRIO"

KEYS_GENERAL = {
        " Espaço ": "Muda entre calendário e diário",
        "   /   ": "Alternar tela dividida",
        "   *   ": "Alternar privacidade global",
        "   ?   ": "Alternar essa ajuda",
        "   Q   ": "Reload",
        "   q   ": "Sair",
        }

KEYS_CALENDAR = {
        "  a(A) ": "Adiciona um evento (recorrente)",
        "   n   ": "Próximo mês (dia)",
        "   p   ": "Mês (dia) anterior",
        "   x   ": "Deletar um evento",
        "   r   ": "Editar um evento",
        "  m(M) ": "Mover evento",
        "   v   ": "Alternar visualização diária/mensal",
        "  g(G) ": "Ir para dia específico",
        "   h   ": "Alternar evento como alta prioridade",
        "   l   ": "Alternar evento como baixa prioridade",
        "   d   ": "Toggle event as done",
        "   .   ": "Alternar privacidade do evento",
        "   C   ": "Importar eventos do calcurse",
        "   G   ": "Retornar para o mês (dia) recorrente",
        }

KEYS_TODO = {
        "  a(A) ": "Adicionar uma nova (sub)tarefa",
        "  h(H) ": "Alternar uma (todas) as tarefas como alta prioridade",
        "  l(L) ": "Alternar uma (todas) as tarefas como baixa prioridade",
        "  d(D) ": "Alternar uma (todas) as tarefas como feitas",
        "  u(U) ": "Desmarcar uma (todas) as tarefas",
        "  x(X) ": "Deletar uma (todas) as tarefas (com todas as subtarefas)",
        "  t(T) ": "Começar/pausar temporizador para a tarefa",
        "   r   ": "Editar uma tarefa",
        "   s   ": "Alternar entre tarefa e subtarefa",
        "   .   ": "Alternar privacidade da tarefa",
        "  f(F) ": "Mudar (remover) data limite da tarefa",
        "   m   ": "Mover tarefa",
        "  C(W) ": "Importar tarefas do calcurse (taskwarrior)",
        }

MSG_NAME          = "CALCURE"
MSG_VIM           = "Teclas Vim (j, k, ZZ, ZQ) também funcionam!"
MSG_INFO          = "Para mais informações, visite:"
MSG_SITE          = "https://anufrievroman.gitbook.io/calcure"
MSG_EXIT          = "Realmente sair? (y/n) "

MSG_EVENT_HIGH    = "Marca como alta prioridade evento número: "
MSG_EVENT_LOW     = "Marca como baixa prioridade evento número: "
MSG_EVENT_DONE    = "Mark as done event number: "
MSG_EVENT_RESET   = "Reseta status para evento número:"
MSG_EVENT_DEL     = "Deleta evento número: "
MSG_EVENT_REN     = "Renomeia evento número: "
MSG_NEW_TITLE     = "Digitar novo título: "
MSG_EVENT_MV      = "Move evento número: "
MSG_EVENT_MV_TO   = "Mover evento para (YYYY/MM/DD): "
MSG_EVENT_MV_TO_D = "Mover evento para: "
MSG_EVENT_DATE    = "Digitar data: "
MSG_EVENT_TITLE   = "Digitar título: "
MSG_EVENT_REP     = "Repetir evento quantas vezes: "
MSG_EVENT_FR      = "Repetir evento cada (d)ia, (w)semana, (m)ês ou (y)ano? "
MSG_EVENT_IMP     = "Importar eventos do Calcurse? (y/n)"
MSG_EVENT_PRIVACY = "Alternar privacidade do evento número: "
MSG_TM_ADD        = "Adicionar/pausar temporizador para tarefa número: "
MSG_TM_RESET      = "Remover temporizador para tarefa número: "
MSG_TS_HIGH       = "Marcar como alta prioridade tarefa número: "
MSG_TS_LOW        = "Marcar como baixa prioridade tarefa número: "
MSG_TS_RES        = "Resetar status para tarefa número: "
MSG_TS_DONE       = "Marcar como feita tarefa número: "
MSG_TS_DEL        = "Deletar tarefa número: "
MSG_TS_DEL_ALL    = "Quer mesmo deletar todas as tarefas? (y/n)"
MSG_TS_EDT_ALL    = "Do you confirm this action? (y/n)"
MSG_TS_MOVE       = "Mover tarefa de número: "
MSG_TS_MOVE_TO    = "Mover tarefa para número: "
MSG_TS_EDIT       = "Editar tarefa número: "
MSG_TS_TOG        = "Alternar subtarefa número: "
MSG_TS_SUB        = "Adicionar subtarefa para tarefa número: "
MSG_TS_TITLE      = "Digitar subtarefa: "
MSG_TS_IM         = "Importar tarefas do Calcurse? (y/n)"
MSG_TS_TW         = "Importar tarefas do Taskwarrior? (y/n)"
MSG_TS_NOTHING    = "Nada planejado..."
MSG_TS_PRIVACY    = "Alternar privacidade de tarefa número: "
MSG_TS_DEAD_ADD   = "Adicionar data limite da tarefa número: "
MSG_TS_DEAD_DEL   = "Remover data limite da tarefa número: "
MSG_TS_DEAD_DATE  = "Adicionar data limite em (AAAA/MM/DD): "
MSG_WEATHER       = "Clima está carregando..."
MSG_ERRORS        = "Errors have occurred. See info.log file in your config folder."
MSG_GOTO          = "Go to date (YYYY/MM/DD): "
MSG_GOTO_D        = "Go to date: "
MSG_INPUT         = "Incorrect input."

CALENDAR_HINT     = "Espaço · Mudar para diário   a · Adicionar evento  n/p · Mudar mês   ? · Todas as combinações de teclas"
CALENDAR_HINT_D   = "Espaço · Mudar para diário   a · Adicionar evento  n/p · Mudar dia   ? · Todas as combinações de teclas"
JOURNAL_HINT      = "Espaço · Mudar para calendar   a · Adicionar tarefa   d · Feito   i · Importante   ? · Todas as combinações de teclas"

DAYS = ["SEGUNDA", "TERÇA", "QUARTA", "QUINTA", "SEXTA", "SÁBADO", "DOMINGO"]
DAYS_PERSIAN = ["SHANBEH", "YEKSHANBEH", "DOSHANBEH", "SESHANBEH", "CHAHARSHANBEH", "PANJSHANBEH", "JOMEH"]

MONTHS = ["JANEIRO", "FEVEREIRO", "MARÇO", "ABRIL", "MAIO", "JUNHO", "JULHO", "AGOSTO", "SETEMBRO", "OUTUBRO", "NOVEMBRO", "DEZEMBRO"]
MONTHS_PERSIAN = ["FARVARDIN", "ORDIBEHESHT", "KHORDAD", "TIR", "MORDAD", "SHAHRIVAR", "MEHR", "ABAN", "AZAR", "DEY", "BAHMAN", "ESFAND"]



"""Chinese translations of the program interface"""

MSG_WELCOME_1 = "欢迎来到 Calcure"
MSG_WELCOME_2 = "你的终端日历和任务管理器!"
MSG_WELCOME_3 = "配置文件和数据文件被创建在:"
MSG_WELCOME_4 = "用于支持、贡献和其他信息，请参看:"
MSG_WELCOME_5 = "按 ? 查看所有的键位绑定或者任何其它的键位以继续。"

TITLE_KEYS_GENERAL  = "通用的键位绑定"
TITLE_KEYS_CALENDAR = "日历栏的键位绑定"
TITLE_KEYS_JOURNAL  = "通知栏的键位绑定"

KEYS_GENERAL = {
        " Space ": "切换日历和通知栏目",
        "   /   ": "切换屏幕分割模式",
        "   *   ": "切换全局隐私模式",
        "   ?   ": "显示此帮助文档",
        "   Q   ": "重新加载",
        "   q   ": "退出",
        }

KEYS_CALENDAR = {
        "  a(A) ": "增加一个（再次发生）事件",
        "   n   ": "下个月 (日)",
        "   p   ": "上个月 (日)",
        "   x   ": "删除事件",
        "   r   ": "编辑事件",
        "  m(M) ": "移动事件",
        "  g(G) ": "跳转到指定的一天",
        "   v   ": "切换每天/每月视图模式",
        "   h   ": "切换事件为更高的优先级",
        "   l   ": "切换事件为更低的优先级",
        "   d   ": "Toggle event as done",
        "   .   ": "切换事件为隐私模式",
        "   C   ": "从calcurse中导入事件",
        "   G   ": "返回当前的月 (日)",
        }

KEYS_TODO = {
        "  a(A) ": "新增一个 (子) 任务",
        "  h(H) ": "切换一个 (所有) 任务为更高优先级",
        "  l(L) ": "切换一个 (所有) 任务为更低优先级",
        "  d(D) ": "切换一个 (所有) 任务为已完成",
        "  u(U) ": "不标记一个 (所有) 任务",
        "  x(X) ": "删除一个 (所有) 任务 (和其子任务)",
        "  t(T) ": "开始/暂停 (移除) 时刻为一个任务",
        "   r   ": "编辑一个任务",
        "   s   ": "切换任务和子任务",
        "   .   ": "切换任务为隐私模式",
        "  f(F) ": "修改 (移除) 任务的截止日期",
        "   m   ": "移动一个任务",
        "  C(W) ": "导入任务从calcurse (taskwarrior)",
        }

MSG_NAME          = "CALCURE"
MSG_VIM           = "Vim 键位 (j, k, ZZ, ZQ) 也可以工作!"
MSG_INFO          = "对于更多信息, 参看:"
MSG_SITE          = "https://anufrievroman.gitbook.io/calcure"
MSG_EXIT          = "是否真的要退出? (y/n) "

MSG_EVENT_HIGH    = "标记为更高优先级的事件号码: "
MSG_EVENT_LOW     = "标记为更低优先级的事件号码: "
MSG_EVENT_DONE    = "Mark as done event number: "
MSG_EVENT_RESET   = "重设状态的事件号码: "
MSG_EVENT_DEL     = "删除事件的号码: "
MSG_EVENT_REN     = "重命名事件的号码: "
MSG_NEW_TITLE     = "输入新的标题: "
MSG_EVENT_MV      = "移动事件的号码: "
MSG_EVENT_MV_TO   = "移动事件到 (YYYY/MM/DD): "
MSG_EVENT_MV_TO_D = "移动事件到: "
MSG_EVENT_DATE    = "输入日期: "
MSG_EVENT_TITLE   = "输入标题: "
MSG_EVENT_REP     = "事件重复多少次: "
MSG_EVENT_FR      = "重复事件 (d)天, (w)周, (m)月 或者 (y)年? "
MSG_EVENT_IMP     = "是否从Calcurse中导入事件? (y/n)"
MSG_EVENT_PRIVACY = "切换为隐私模式的事件号码: "
MSG_TM_ADD        = "增加/暂停计时器的任务号码: "
MSG_TM_RESET      = "移除计时器的任务号码: "
MSG_TS_HIGH       = "标记为更高优先级的任务号码: "
MSG_TS_LOW        = "标记为更低优先级的任务号码: "
MSG_TS_RES        = "重设状态的任务号码: "
MSG_TS_DONE       = "标记为已完成的任务号码: "
MSG_TS_DEL        = "删除的任务号码: "
MSG_TS_DEL_ALL    = "是否真的删除所有任务? (y/n)"
MSG_TS_EDT_ALL    = "Do you confirm this action? (y/n)"
MSG_TS_MOVE       = "移动任务的号码: "
MSG_TS_MOVE_TO    = "移动任务到新的号码: "
MSG_TS_EDIT       = "编辑任务的号码: "
MSG_TS_TOG        = "切换子任务的号码: "
MSG_TS_SUB        = "添加子任务的任务号码: "
MSG_TS_TITLE      = "输入子任务: "
MSG_TS_IM         = "是否从Calcurse导入任务? (y/n)"
MSG_TS_TW         = "是否从Taskwarrior导入任务? (y/n)"
MSG_TS_NOTHING    = "没有事情被计划..."
MSG_TS_PRIVACY    = "切换为隐私模式的事件号码: "
MSG_TS_DEAD_ADD   = "增加截至日期为事件号码: "
MSG_TS_DEAD_DEL   = "移除截止日期为事件号码: "
MSG_TS_DEAD_DATE  = "增加截至日期在(YYYY/MM/DD): "
MSG_WEATHER       = "天气插件正在加载..."
MSG_ERRORS        = "Errors have occurred. See info.log in your config folder."
MSG_GOTO          = "Go to date (YYYY/MM/DD): "
MSG_INPUT         = "Incorrect input."

CALENDAR_HINT     = "Space · 转换到通知栏   a · 增加事件  n/p · 改变月   ? · 所有键位绑定"
CALENDAR_HINT_D   = "Space · 转换到通知栏   a · 增加事件  n/p · 改变日   ? · 所有键位绑定"
JOURNAL_HINT      = "Space · 转换到日历栏   a · 增加任务   v · 已完成   i · 重要的   ? · 所有键位绑定"

DAYS = ["周一", "周二", "周三", "周四", "周五", "周六", "周日"]
DAYS_PERSIAN = ["SHANBEH", "YEKSHANBEH", "DOSHANBEH", "SESHANBEH", "CHAHARSHANBEH", "PANJSHANBEH", "JOMEH"]

MONTHS = ["一月", "二月", "三月", "四月", "五月", "六月", "七月", "八月", "九月", "十月", "十一月", "十二月"]
MONTHS_PERSIAN = ["FARVARDIN", "ORDIBEHESHT", "KHORDAD", "TIR", "MORDAD", "SHAHRIVAR", "MEHR", "ABAN", "AZAR", "DEY", "BAHMAN", "ESFAND"]



"""Slovak translations of the program interface"""

MSG_WELCOME_1 = "Víta vás Calcure"
MSG_WELCOME_2 = "Váš kalendár a správca úloh pre terminál!"
MSG_WELCOME_3 = "Súbory s dátami a konfiguráciou boli vytvorené v:"
MSG_WELCOME_4 = "Ak potrebujete viac informácií o podpore, prispievaní atď, navštívte:"
MSG_WELCOME_5 = "Stlačte ? na zobrazenie všetkých klávesových skratiek alebo stlačte akýkoľvek iný kláves na pokračovanie."

TITLE_KEYS_GENERAL = "SPOLOČNÉ SKRATKY"
TITLE_KEYS_CALENDAR = "SKRATKY PRE KALENDÁR"
TITLE_KEYS_JOURNAL  = "SKRATKY PRE DENNÍK"

KEYS_GENERAL = {
        " Medz ": "Prepnúť medzi kalendárom a denníkom",
        "   /   ": "Vypnúť/zapnúť rozdelenie obrazovky",
        "   *   ": "Vypnúť/zapnúť globálne súkromie",
        "   ?   ": "Vypnúť/zapnúť tohto pomocníka",
        "   Q   ": "Reload",
        "   q   ": "Ukončiť",
        }

KEYS_CALENDAR = {
        "  a(A) ": "Pridať (opakujúcu sa) udalosť",
        "   n   ": "Nasledujúci mesiac (deň)",
        "   p   ": "Predchádzajúci mesiac (deň)",
        "   x   ": "Odstrániť udalosť",
        "   r   ": "Premenovať udalosť",
        "  m(M) ": "Presunúť udalosť (v rámci mesiaca)",
        "  g(G) ": "Prejsť na určitý deň (v rámci mesiaca)",
        "   v   ": "Prepnúť medzi zobrazením dní/mesiacov",
        "   h   ": "Nastaviť/zrušiť vysokú prioritu pre udalosť",
        "   l   ": "Nastaviť/zrušiť nízku prioritu pre udalosť",
        "   d   ": "Označiť udalosť ako (ne)dokončenú",
        "   .   ": "Označiť udalosť ako súkromnú/verejnú",
        "   C   ": "Importovať udalosti z calcurse",
        "   R   ": "Vrátiť sa na aktuálny mesiac (deň)",
        }

KEYS_TODO = {
        "  a(A) ": "Pridať novú (pod)úlohu",
        "  h(H) ": "Nastaviť/zrušiť vysokú prioritu pre jednu (všetky) úlohu",
        "  l(L) ": "Nastaviť/zrušiť nízku prioritu pre jednu (všetky) úlohu",
        "  d(D) ": "Označiť jednu (všetky) úlohu ako (ne)dokončenú",
        "  u(U) ": "Zrušiť označenie jednej (všetkých) úlohy",
        "  x(X) ": "Odstrániť jednu (všetky) úlohy (aj s podúlohami)",
        "  t(T) ": "Spustiť/pozastaviť (odstrániť) časovač úlohy",
        "   r   ": "Premenovať úlohu",
        "   s   ": "Prepnúť medzi úlohou a podúlohou",
        "   .   ": "Prepnúť viditeľnosť úlohy",
        "  f(F) ": "Zmeniť (odstrániť) hraničný termín úlohy",
        "   m   ": "Odstrániť úlohu",
        "  C(W) ": "Importovať úlohy z calcurse (taskwarrior)",
        }

MSG_NAME          = "CALCURE"
MSG_VIM           = "Fungujú aj Vim klávesy (j, k, ZZ, ZQ)!"
MSG_INFO          = "Pre viac informácií navštívte:"
MSG_SITE          = "https://anufrievroman.gitbook.io/calcure"
MSG_EXIT          = "Naozaj ukončiť? (y/n) "

MSG_EVENT_HIGH    = "Nastaviť vysokú prioritu udalosti číslo: "
MSG_EVENT_LOW     = "Nastaviť nízku prioritu udalosti číslo: "
MSG_EVENT_DONE    = "Označiť ako hotovú udalosť číslo: "
MSG_EVENT_RESET   = "Obnoviť stav udalosti číslo: "
MSG_EVENT_DEL     = "Odstrániť udalosť číslo: "
MSG_EVENT_REN     = "Premenovať udalosť číslo: "
MSG_NEW_TITLE     = "Zadať nový názov: "
MSG_EVENT_MV      = "Presunúť udalosť číslo: "
MSG_EVENT_MV_TO   = "Presunúť udalosť na (RRRR/MM/DD): "
MSG_EVENT_MV_TO_D = "Presunúť udalosť na: "
MSG_EVENT_DATE    = "Zadajte dátum: "
MSG_EVENT_TITLE   = "Zadajte názov: "
MSG_EVENT_REP     = "Koľkokrát sa udalosť opakuje: "
MSG_EVENT_FR      = "Opakovať udalosť každý (d)eň, týždeň (w), (m)esiac alebo rok (y)? "
MSG_EVENT_IMP     = "Importovať udalosti z Calcurse? (y/n)"
MSG_EVENT_PRIVACY = "Prepnúť viditeľnosť úlohy číslo: "
MSG_TM_ADD        = "Pridať/pozastaviť časovač úlohy číslo: "
MSG_TM_RESET      = "Odstrániť časovač úlohy číslo: "
MSG_TS_HIGH       = "Nastaviť vysokú prioritu úlohe číslo: "
MSG_TS_LOW        = "Nastaviť nízku prioritu úlohe číslo: "
MSG_TS_RES        = "Obnoviť stav úlohy číslo: "
MSG_TS_DONE       = "Označiť ako hotovú úlohu číslo: "
MSG_TS_DEL        = "Odstrániť úlohu číslo: "
MSG_TS_DEL_ALL    = "Naozaj odstrániť všetky úlohy? (y/n)"
MSG_TS_EDT_ALL    = "Do you confirm this action? (y/n)"
MSG_TS_MOVE       = "Presunúť úlohu z čísla: "
MSG_TS_MOVE_TO    = "Presunúť úlohu na číslo: "
MSG_TS_EDIT       = "Upraviť úlohu číslo: "
MSG_TS_TOG        = "Vypnúť/zapnúť podúlohu číslo: "
MSG_TS_SUB        = "Pridať podúlohu úlohe číslo: "
MSG_TS_TITLE      = "Zadajte podúlohu: "
MSG_TS_IM         = "Importovať úlohy z Calcurse? (y/n)"
MSG_TS_TW         = "Importovať úlohy z Taskwarriora? (y/n)"
MSG_TS_NOTHING    = "Nič nie je naplánované..."
MSG_TS_PRIVACY    = "Vypnúť/zapnúť viditeľnosť úlohy číslo: "
MSG_TS_DEAD_ADD   = "Pridať hraničný termín úlohy číslo: "
MSG_TS_DEAD_DEL   = "Odstrániť hraničný termín ulohy číslo: "
MSG_TS_DEAD_DATE  = "Pridať hraničný termín dňa (RRRR/MM/DD): "
MSG_WEATHER       = "Počasie sa načíta..."
MSG_ERRORS        = "Došlo k chybe. Otvorte info.log v priečinku s konfiguráciou."
MSG_INPUT         = "Neplatný vstup."
MSG_GOTO          = "Prejsť na dátum (RRRR/MM/DD): "
MSG_GOTO_D        = "Prejsť na dátum: "

CALENDAR_HINT     = "Medzera · Prepnúť na denník   a · Pridať udalosť  n/p · Zmeniť mesiac ? · Všetky skratky"
CALENDAR_HINT_D   = "Medzera · Prepnúť na denník   a · Pridať udalosť  n/p · Zmeniť deň   ? · Všetky skratky"
JOURNAL_HINT      = "Medzera · Prepnúť na kalendár   a · Pridať úlohu d · Hotovo   i · Dôležité   ? · Všetky skratky"

DAYS = ["PONDELOK", "UTOROK", "STREDA", "ŠTVRTOK", "PIATOK", "SOBOTA", "NEDEĽA"]
DAYS_PERSIAN = ["SHANBEH", "YEKSHANBEH", "DOSHANBEH", "SESHANBEH", "CHAHARSHANBEH", "PANJSHANBEH", "JOMEH"]

MONTHS = ["JANUÁR", "FEBRUÁR", "MAREC", "APRÍL", "MÁJ", "JÚN", "JÚL", "AUGUST", "SEPTEMBER", "OKTÓBER", "NOVEMBER", "DECEMBER"]
MONTHS_PERSIAN = ["FARVARDIN", "ORDIBEHESHT", "KHORDAD", "TIR", "MORDAD", "SHAHRIVAR", "MEHR", "ABAN", "AZAR", "DEY", "BAHMAN", "ESFAND"]



"""Module that controls loading data from files and libraries"""

import configparser
import csv
import os
import datetime
import icalendar
import urllib.request
import io
import logging

from pathlib import Path

from calcure.data import *
from calcure.calendars import convert_to_persian_date


class LoaderCSV:
    """Load data from CSV files"""

    def create_file(self, filename):
        """Create CSV file"""
        try:
            with open(filename, "w+", encoding="utf-8") as file:
                pass
            return []
        except (FileNotFoundError, NameError) as e_message:
            logging.error("Problem occured trying to create %s. %s", filename, e_message)
            return []

    def read_file(self, filename):
        """Read CSV file or create new one if it does not exist"""
        try:
            with open(filename, "r", encoding="utf-8") as file:
                lines = csv.reader(file, delimiter = ',')
                return list(lines)
        except IOError: # File does not exist
            # logging.info("Creating %s.", filename)
            return self.create_file(filename)


class TaskLoaderCSV(LoaderCSV):
    """Load tasks from CSV files"""

    def __init__(self, cf):
        self.user_tasks = Tasks()
        self.tasks_file = cf.TASKS_FILE
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    @property
    def is_task_format_old(self):
        """Check if the database format is old"""
        with open(self.tasks_file, "r", encoding="utf-8") as f:
            text = f.read()
        return text[0] == '"'

    def load(self):
        """Reads from CSV file"""

        self.user_tasks.delete_all_items()
        lines = self.read_file(self.tasks_file)

        for index, row in enumerate(lines):
            task_id = index

            # Read task dates:
            if self.is_task_format_old:
                shift = 0
                year = 0
                month = 0
                day = 0
            else:
                shift = 3
                year = int(row[0])
                month = int(row[1])
                day = int(row[2])

            # Convert to persian date if needed and if it is not zero date:
            if self.use_persian_calendar and year != 0:
                year, month, day = convert_to_persian_date(year, month, day)

            # Read task name and statuses:
            if row[0 + shift][0] == '.':
                name = row[0 + shift][1:]
                is_private = True
            else:
                name = row[0 + shift]
                is_private = False
            status = Status[row[1 + shift].upper()]
            stamps = row[(2 + shift):] if len(row) > 2 else []
            timer = Timer(stamps)

            # Add task:
            new_task = Task(task_id, name, status, timer, is_private, year, month, day)
            self.user_tasks.add_item(new_task)
        self.user_tasks.changed = False
        return self.user_tasks


class EventLoaderCSV(LoaderCSV):
    """Load events from CSV files"""

    def __init__(self, cf):
        self.user_events = Events()
        self.events_file = cf.EVENTS_FILE
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def load(self):
        """Read from CSV file"""
        self.user_events.delete_all_items()
        lines = self.read_file(self.events_file)
        for index, row in enumerate(lines):
            event_id = index
            year = int(row[1])
            month = int(row[2])
            day = int(row[3])
            if row[4][0] == '.':
                name = row[4][1:]
                is_private = True
            else:
                name = row[4]
                is_private = False

            # Account for old versions of the datafile:
            if len(row) > 5:
                repetition = int(row[5])
                if row[6] == 'd':
                    frequency = Frequency.DAILY
                elif row[6] == 'w':
                    frequency = Frequency.WEEKLY
                elif row[6] == 'm':
                    frequency = Frequency.MONTHLY
                elif row[6] == 'y':
                    frequency = Frequency.YEARLY
                else:
                    try:
                        frequency = Frequency[row[6].upper()]
                    except (ValueError, KeyError):
                        frequency = Frequency.ONCE
            else:
                repetition = '1'
                frequency = Frequency.ONCE
            if len(row) > 7:
                status = Status[row[7].upper()]
            else:
                status = Status.NORMAL

            # Convert to persian date if needed:
            if self.use_persian_calendar:
                year, month, day = convert_to_persian_date(year, month, day)

            # Add event:
            new_event = UserEvent(event_id, year, month, day, name, repetition, frequency, status, is_private)
            self.user_events.add_item(new_event)
        self.user_events.changed = False
        return self.user_events


class HolidayLoader:
    """Load holidays for this country around this year"""

    def __init__(self, cf):
        self.holidays = Events()
        self.countries = cf.HOLIDAY_COUNTRY.split(',')
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def load(self):
        """Run and collect holidays for each country"""
        holidays = Events()
        for country in self.countries:
            self.load_country(country)
        return self.holidays

    def load_country(self, country):
        """Load list of holidays from 'holidays' module"""

        def get_country_and_subdivision(country):
            """Get country and subdivision, where encoded."""
            if ":" in country:
                return tuple(country.split(":"))
            else:
                return (country, None)

        try:
            import holidays as hl
            from holidays import registry

            (country, subdivision) = get_country_and_subdivision(country)
            country_codes = {x[0]: x[2] for x in registry.COUNTRIES.values()}
            country_code = country_codes.get(country)
            year = datetime.date.today().year
            holiday_events = (getattr(hl, country))(subdiv=subdivision, years=[year+x for x in range(-2, 5)])
            for date, name in holiday_events.items():

                # Convert to persian date if needed:
                if self.use_persian_calendar:
                    year, month, day = convert_to_persian_date(date.year, date.month, date.day)
                else:
                    year, month, day = date.year, date.month, date.day

                # Add holiday:
                holiday = Event(year, month, day, f'{name} ({country_code})' if len(self.countries) > 1 else name)
                self.holidays.add_item(holiday)

        except ModuleNotFoundError:
            logging.error("Couldn't load holidays. Module holidays is not installed. Try 'pip install holidays'")
            pass
        except (SyntaxError, AttributeError) as e_message:
            logging.error("Couldn't load holidays. Country might be incorrect. %s", e_message)
            pass
        return self.holidays


class BirthdayLoader:
    """Load birthdays of contacts"""

    def __init__(self, cf):
        self.birthdays = Birthdays()
        self.abook_file = Path.home() / ".abook" / "addressbook"
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR
        self.load_birthdays = cf.BIRTHDAYS_FROM_ABOOK

    def load(self):
        """Loading birthdays from abook contacts"""

        # Quit if birthdays do not need to be loaded:
        if not self.load_birthdays:
            return self.birthdays

        # Quit if file does not exists:
        if not self.abook_file.exists():
            logging.warning("Couldn't load birthdays. File. %s does not exist.", str(self.abook_file))
            return self.birthdays

        abook = configparser.ConfigParser()
        abook.read(self.abook_file)
        for each_contact in abook.sections():
            for key, _ in abook.items(each_contact):
                if key in ["birthday", "anniversary"]:
                    month = int(abook[each_contact][key][-5:-3])
                    day = int(abook[each_contact][key][-2:])
                    name = abook[each_contact]["name"]

                    # Convert to persian date if needed:
                    if self.use_persian_calendar:
                        _, month, day = convert_to_persian_date(1000, month, day)

                    # Add birthday:
                    birthday = Event(1, month, day, name)
                    self.birthdays.add_item(birthday)
        return self.birthdays


class LoaderICS:
    """Load data from ICS files"""

    def read_lines(self, file):
        """Read the file line-by-line and remove multiple PRODID lines"""
        already_has_prodid = False
        text = io.StringIO()
        for line in file:
            # If there is more than one PRODID line or a TZUNTIL line, skip them:
            if ((not (already_has_prodid and "PRODID:" in line)) and
                (not "TZUNTIL" in line)):
                text.write(line)

            if "PRODID:" in line:
                already_has_prodid = True
        return text.getvalue()

    def read_file(self, path):
        """Parse an ics file if it exists"""
        if not os.path.exists(path):
            logging.error("Failed to load %s because file does not exist.", path)
            return ""
        with open(path, 'r', encoding="utf-8") as file:
            return self.read_lines(file)

    def read_url(self, path):
        """Parse an ics URL if it exists and networks works"""
        try:
            with urllib.request.urlopen(path) as response:
                file = io.TextIOWrapper(response, 'utf-8')
                return self.read_lines(file)
        except urllib.error.HTTPError as e_message:
            logging.error("Failed to load from %s. Probably url is wrong. %s", path, e_message)
            return ""
        except urllib.error.URLError as e_message:
            logging.error("Failed to load from %s. Probably no internet connection. %s", path, e_message)
            return ""

    def read_resource(self, path):
        """Determine type of the resource, parse it, and return list of strings for each file"""
        ics_files = []
        path = os.path.expanduser(path)

        # If it's a URL, try to load it:
        if path.startswith('http'):
            ics_files.append(self.read_url(path))
            return ics_files

        # If it's a local file, read it:
        if path.endswith('.ics'):
            ics_files.append(self.read_file(path))
            return ics_files

        # Otherwise, assume it's a folder, and read every file inside:
        for root, directories, files in os.walk(path):
            for filename in files:
                # Get the full path to the file
                file_path = os.path.join(root, filename)

                # `path` may contain files with metadata, e.g. `color` and
                # `displayname`. For now, exclude those while loading.
                if file_path.endswith('.ics'):
                    ics_files.append(self.read_file(file_path))

        return ics_files


class TaskLoaderICS(LoaderICS):
    """Load tasks from ICS files"""

    def __init__(self, cf):
        self.user_ics_tasks = Tasks()
        self.ics_task_files = cf.ICS_TASK_FILES
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR

    def parse_task(self, component, calendar_number):
        """Parse single task and add it to the user_ics_tasks"""
        task_status = component.get('status')

        if task_status == "CANCELLED":
            return

        task_priority = component.get('priority')
        task_name = str(component.get('summary'))
        task_id = self.user_ics_tasks.generate_id()

        # Assign status from priority:
        status = Status.NORMAL
        if task_priority is not None:
            if task_priority > 5:
                status = Status.UNIMPORTANT
            if task_priority < 5:
                status = Status.IMPORTANT

        # Correct according to status:
        if task_status == "COMPLETED":
            status = Status.DONE

        # Try reading task due date:
        due_dt = None
        try:
            due_dt = component.get('due').dt
            year, month, day = due_dt.year, due_dt.month, due_dt.day
        except AttributeError:
            year, month, day = 0, 0, 0

        timer = Timer([])
        is_private = False

        # Add task:
        new_task = Task(task_id, task_name, status, timer, is_private,
                        year, month, day, calendar_number)
        self.user_ics_tasks.add_item(new_task)


    def load(self):
        """Load tasks from each of the ics files"""

        # Quit if the files are not specified in config:
        if self.ics_task_files is None:
            return self.user_ics_tasks

        self.user_ics_tasks.delete_all_items()
        for calendar_number, filename in enumerate(self.ics_task_files):
            # For each resource from config, load a list that has one or more ics files:
            ics_files = self.read_resource(filename)
            for ics_file in ics_files:
                try:
                    cal = icalendar.Calendar.from_ical(ics_file)
                    for component in cal.walk():
                        if component.name == 'VTODO':
                            self.parse_task(component, calendar_number)
                except Exception as e_message:
                    logging.error("Failed to parse %s. %s", filename, e_message)

        return self.user_ics_tasks


class EventLoaderICS(LoaderICS):
    """Load events from ICS files"""

    def __init__(self, cf):
        self.user_ics_events = Events()
        self.ics_event_files = cf.ICS_EVENT_FILES
        self.use_persian_calendar = cf.USE_PERSIAN_CALENDAR
        self.local_timezone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

    def parse_event(self, component, index, calendar_number):
        """Parse single event and add it to user_ics_events"""

        # Default parameters:
        hour = None
        minute = None
        event_id = index
        repetition = 1
        frequency = Frequency.ONCE
        status = Status.NORMAL
        is_private = False
        rrule = None
        exdate = None

        # Parameters of the event from ics file, if they exist:
        name = str(component.get('summary', ''))
        dt = None
        try:
            dt = component.get('dtstart').dt

            # Convert to local timezone if it's provided:
            if hasattr(dt, "tzinfo"):
                dt = dt.astimezone(self.local_timezone)

            year, month, day = dt.year, dt.month, dt.day
        except AttributeError:
            year, month, day = 0, 1, 1

        # See if this event takes multiple days:
        try:
            dt_end = component.get('dtend').dt

            # Convert to local timezone if it's provided:
            if hasattr(dt_end, "tzinfo"):
                dt_end = dt_end.astimezone(self.local_timezone)

            # For events with time:
            try:
                dt_difference = dt_end.date() - dt.date()
                if dt_difference.days > 0:
                    repetition = dt_difference.days + 1
                    frequency = Frequency.DAILY

            # For all day events, last day does not count:
            except:
                dt_difference = dt_end - dt
                if dt_difference.days > 0:
                    repetition = dt_difference.days
                    frequency = Frequency.DAILY
            
            # Parsing recuring rules:
            if 'rrule' in component:
                rrule = component.get('rrule').to_ical().decode('utf-8')
                exdate = component.get('exdate')
                repetition = 0

        except AttributeError:
            logging.error("Failed to parse event %s on %s.", name, dt)
            pass

        # Add start time to non-all-day events:
        all_day = component.get('dtstart').params.get('VALUE') == 'DATE' if component.get('dtstart') else False
        if not all_day:
            hour = dt.hour if dt else 0
            minute = dt.minute if dt else 0

        # Convert to persian date if needed:
        if self.use_persian_calendar:
            year, month, day = convert_to_persian_date(year, month, day)

        # Add event:
        new_event = UserEvent(event_id, year, month, day, name, repetition, frequency,
                              status, is_private, calendar_number, hour, minute, rrule, exdate)
        self.user_ics_events.add_item(new_event)

    def load(self):
        """Load events from each of the ics files"""

        # Quit if the files are not specified in config:
        if self.ics_event_files is None:
            return self.user_ics_events

        self.user_ics_events.delete_all_items()
        for calendar_number, filename in enumerate(self.ics_event_files):

            # For each resource from config, load a list that has one or more ics files:
            ics_files = self.read_resource(filename)
            for ics_file in ics_files:
                try:
                    cal = icalendar.Calendar.from_ical(ics_file)
                    index = 0
                    for component in cal.walk():
                        if component.name == 'VEVENT':
                            index += 1
                            self.parse_event(component, index, calendar_number)

                except Exception as e_message:
                    logging.error("Failed to parse %s. %s", filename, e_message)
        return self.user_ics_events



#!/usr/bin/env python

"""This is the main module that contains views and the main logic"""

import curses
import time
import getopt
import sys
import importlib
import threading

from calcure.calendars import Calendar
from calcure.errors import Error
from calcure.configuration import Config
from calcure.weather import Weather
from calcure.importers import Importer
from calcure.dialogues import clear_line
from calcure.screen import Screen
from calcure.savers import TaskSaverCSV, EventSaverCSV
from calcure.colors import Color, initialize_colors
from calcure.loaders import *
from calcure.data import *
from calcure.controls import *


# Initialise config:
cf = Config()
error = Error(cf.config_folder / "info.log")

# Language:
if cf.LANG == "fr":
    from calcure.translations.fr import *
elif cf.LANG == "ru":
    from calcure.translations.ru import *
elif cf.LANG == "it":
    from calcure.translations.it import *
elif cf.LANG == "br":
    from calcure.translations.br import *
elif cf.LANG == "tr":
    from calcure.translations.tr import *
elif cf.LANG == "zh":
    from calcure.translations.zh import *
elif cf.LANG == "sk":
    from calcure.translations.sk import *
else:
    from calcure.translations.en import *


__version__ = "3.1"


def read_items_from_user_arguments(screen, user_tasks, user_events, task_saver_csv, event_saver_csv):
    """Read --task and --event flags from user arguments to create new tasks or events"""
    try:
        opts, _ = getopt.getopt(sys.argv[1:], "pjhvi", ["folder=", "config=", "task=", "event="])
        for opt, arg in opts:
            if opt in '--task':
                name = arg
                user_tasks.add_item(Task(len(user_tasks.items), name, Status.NORMAL, Timer([]), False))
                screen.state = AppState.EXIT
                task_saver_csv.save()
            if opt in '--event':
                year = int(arg.split("-")[0])
                month = int(arg.split("-")[1])
                day = int(arg.split("-")[2])
                name = arg.split("-")[3]
                event_id = user_events.items[-1].item_id + 1 if not user_events.is_empty() else 1
                user_events.add_item(UserEvent(event_id, year, month, day, name,
                                        1, Frequency.ONCE, Status.NORMAL, False))
                screen.state = AppState.EXIT
                event_saver_csv.save()
    except (getopt.GetoptError, ValueError):
        pass


class View:
    """Parent class of a view that displays things at certain coordinates"""

    def __init__(self, stdscr, y, x):
        self.stdscr = stdscr
        self.y = y
        self.x = x

    def fill_background(self):
        """Fill the screen background with background color"""
        y_max, x_max = self.stdscr.getmaxyx()
        for index in range(y_max - 1):
            self.stdscr.addstr(index, 0, " " * x_max, curses.color_pair(Color.EMPTY.value))

    def display_line(self, y, x, text, color, bold=False, underlined=False):
        """Display the line of text respecting the slyling and available space"""

        # Make sure that we display inside the screen:
        y_max, x_max = self.stdscr.getmaxyx()
        if y >= y_max or x >= x_max:
            return

        # Cut the text if it does not fit the screen:
        real_text = text.replace('\u0336', "")
        number_of_characters = len(real_text)
        avaliable_space = x_max - x
        number_of_special = text.count('\u0336')
        if number_of_characters > avaliable_space:
            coefficient = 2 if number_of_special > 0 else 1
            text = f"{text[:(avaliable_space - 1)*coefficient]}"

        try:
            if bold and underlined:
                self.stdscr.addstr(y, x, text, curses.color_pair(color.value) | curses.A_BOLD | curses.A_UNDERLINE)
            elif bold and not underlined:
                self.stdscr.addstr(y, x, text, curses.color_pair(color.value) | curses.A_BOLD)
            elif underlined and not bold:
                self.stdscr.addstr(y, x, text, curses.color_pair(color.value) | curses.A_UNDERLINE)
            else:
                self.stdscr.addstr(y, x, text, curses.color_pair(color.value))
        except curses.error: # Fix for occasional error with large zoom (reason is unclear)
            return


class TaskView(View):
    """Display a single task"""

    def __init__(self, stdscr, y, x, task, screen):
        super().__init__(stdscr, y, x)
        self.task = task
        self.screen = screen

    @property
    def color(self):
        """Select the color depending on the status"""
        if self.task.status == Status.DONE:
            return Color.DONE
        if self.task.status == Status.IMPORTANT:
            return Color.IMPORTANT
        if self.task.status == Status.UNIMPORTANT:
            return Color.UNIMPORTANT
        return Color.TODO

    @property
    def icon(self):
        """Select the icon for the task"""
        icon = cf.TODO_ICON
        if cf.DISPLAY_ICONS:
            for keyword in cf.ICONS:
                if keyword in self.task.name.lower():
                    icon = cf.ICONS[keyword]
        if self.task.status == Status.DONE:
            icon = cf.DONE_ICON
        if self.task.status == Status.IMPORTANT:
            icon = cf.IMPORTANT_ICON
        return icon

    @property
    def indent(self):
        """Calculate the left indentation depending on the task level"""
        if self.task.name[:4] == '----':
            return 4
        if self.task.name[:2] == '--':
            return 2
        return 0

    @property
    def info(self):
        """Icon and name of the task, which is decorated if needed"""
        name = self.task.name[self.indent:]
        if self.screen.privacy or self.task.privacy:
            return f'{cf.TODO_ICON} {cf.PRIVACY_ICON * len(name)}'
        if self.task.status == Status.DONE and cf.STRIKETHROUGH_DONE:
            strike = "\u0336"
            return f'{self.icon} {strike}{strike.join(name)}{strike}'
        return f'{self.icon} {name}'

    def render(self):
        """Render a line with an icon, task, deadline, and timer"""
        self.display_line(self.y, self.x + self.indent, self.info, self.color)

        deadline_indentation = self.screen.x_min + 2 + len(self.info) + self.indent
        deadline_view = TaskDeadlineView(self.stdscr, self.y, deadline_indentation, self.task)
        deadline_view.render()

        addition_indentation = (deadline_view.has_deadline)*(4 + len(deadline_view.info))
        timer_indentation = deadline_indentation + addition_indentation
        timer_view = TimerView(self.stdscr, self.y, timer_indentation, self.task.timer)
        timer_view.render()


class TaskDeadlineView(View):
    """Display deadline for a task"""

    def __init__(self, stdscr, y, x, task):
        super().__init__(stdscr, y, x)
        self.task = task
        self.color = Color.DEADLINES
        self.icon = cf.DEADLINE_ICON
        self.info = f"{self.task.year}/{self.task.month}/{self.task.day}"
        self.has_deadline = (self.task.year > 0)

    def render(self):
        """Render a line with the deadline date and icon"""
        if self.has_deadline:
            self.display_line(self.y, self.x, f"{self.icon} {self.info}", self.color)


class TimerView(View):
    """Display timer for a task"""

    def __init__(self, stdscr, y, x, timer):
        super().__init__(stdscr, y, x)
        self.timer = timer
        self.color = Color.TIMER if self.timer.is_counting else Color.TIMER_PAUSED

    @property
    def icon(self):
        """Return icon corresponding to timer state"""
        TIMER_RUNS_ICON = "⏵" if cf.DISPLAY_ICONS else "·"
        TIMER_PAUSED_ICON = "⏯︎" if cf.DISPLAY_ICONS else "·"
        return TIMER_RUNS_ICON if self.timer.is_counting else TIMER_PAUSED_ICON

    def render(self):
        """Render a line with a timer and icon"""
        if self.timer.is_started:
            self.display_line(self.y, self.x, f"{self.icon} {self.timer.passed_time}", self.color)


class JournalView(View):
    """Displays a list of all tasks"""

    def __init__(self, stdscr, y, x, user_tasks, user_ics_tasks, screen):
        super().__init__(stdscr, y, x)
        self.user_tasks = user_tasks
        self.user_ics_tasks = user_ics_tasks
        self.screen = screen

    def render(self):
        """Render the list of tasks"""
        if not self.user_tasks.items and not self.user_ics_tasks.items and cf.SHOW_NOTHING_PLANNED:
            self.display_line(self.y, self.x, MSG_TS_NOTHING, Color.UNIMPORTANT)
        for index, task in enumerate(self.user_tasks.items):
            task_view = TaskView(self.stdscr, self.y, self.x, task, self.screen)
            task_view.render()
            if self.screen.selection_mode and self.screen.state == AppState.JOURNAL:
                self.display_line(self.y, self.x, str(index + 1), Color.ACTIVE_PANE)
            self.y += 1

        self.y += 1
        for index, task in enumerate(self.user_ics_tasks.items):
            task_view = TaskView(self.stdscr, self.y, self.x, task, self.screen)
            task_view.render()
            self.y += 1


class EventView(View):
    """Parent class to display events"""

    def __init__(self, stdscr, y, x, event, screen):
        super().__init__(stdscr, y, x)
        self.event = event
        self.screen = screen
        self.start_time = ""

        # Show time for ICS events if it set:
        if hasattr(self.event, 'hour'):
            if self.event.hour is not None:
                self.start_time = f"{self.event.hour:0=2}:{self.event.minute:0=2} "

        self.info = f"{self.icon} {self.start_time}{self.event.name}"

    @property
    def icon(self):
        """Select the right icon for the event"""
        return cf.EVENT_ICON

    @property
    def color(self):
        """Assign color depending on the status or calendar number if it's from .ics file"""
        if self.event.calendar_number is None:
            if self.event.status == Status.DONE:
                return Color.DONE
            if self.event.status == Status.IMPORTANT:
                return Color.IMPORTANT
            if self.event.status == Status.UNIMPORTANT:
                return Color.UNIMPORTANT
            return Color.EVENTS
        else:
            for color in Color:
                if color.value == Color.ICS_CALENDARS0.value + self.event.calendar_number:
                    return color
            return Color.EVENTS

    def decorate_info(self):
        """Icon and name of the event, which is decorated if needed"""
        if self.event.status == Status.DONE and cf.STRIKETHROUGH_DONE:
            strike = "\u0336"
            self.info = f'{self.icon} {strike}{strike.join(self.event.name)}{strike}'

    def obfuscate_info(self):
        """Obfuscate the info if privacy mode is on"""
        if self.screen.privacy or self.event.privacy:
            self.info = f'{cf.EVENT_ICON} {cf.PRIVACY_ICON * len(self.event.name)}'

    def cut_info(self):
        """Cut the name to fit into the cell of the calendar"""
        avaliable_space = self.screen.x_max - self.x
        number_of_special = self.info[:avaliable_space].count('\u0336')
        number_of_special += 2 if number_of_special > 2 else 0
        self.info = self.info[:self.screen.x_max - self.x + number_of_special]
        x_cell = self.screen.x_max // 7
        if (cf.CUT_TITLES or cf.SHOW_CALENDAR_BORDERS) and self.screen.calendar_state == CalState.MONTHLY:
            self.info = self.info[:(x_cell - 1)]

    def minimize_info(self):
        """Reduce the info to just icon if not much space if available"""
        x_cell = self.screen.x_max // 7
        if x_cell < 7:
            self.info = self.icon

    def fill_remaining_space(self):
        """Fill rest of the line of the calendar with empty characters"""
        x_cell = self.screen.x_max // 7
        if len(self.info) < self.screen.x_max - self.x:
            self.info += " "*(self.screen.x_max - self.x - len(self.info))


class UserEventView(EventView):
    """Display a single user event"""

    @property
    def icon(self):
        """Select the right icon for the event"""
        icon = cf.EVENT_ICON
        if cf.DISPLAY_ICONS:
            for keyword in cf.ICONS:
                if keyword in self.event.name.lower():
                    icon = cf.ICONS[keyword]
        if self.screen.privacy or self.event.privacy:
            icon = cf.PRIVACY_ICON
        return icon

    def render(self):
        """Render this view on the screen"""
        self.decorate_info()
        self.obfuscate_info()
        self.fill_remaining_space()
        self.cut_info()
        self.minimize_info()
        self.display_line(self.y, self.x, self.info, self.color)


class BirthdayView(EventView):
    """Display a line with birthday icon and name"""

    @property
    def icon(self):
        """Set the icon for birthdays"""
        return cf.BIRTHDAY_ICON

    def render(self):
        """Render this view on the screen"""
        if self.screen.privacy:
            self.obfuscate_info()
        self.fill_remaining_space()
        self.cut_info()
        self.minimize_info()
        self.display_line(self.y, self.x, self.info, Color.BIRTHDAYS)


class HolidayView(EventView):
    """Display a line with holiday icon and occasion"""

    @property
    def icon(self):
        """Set the icon for holiday events"""
        return cf.HOLIDAY_ICON

    def render(self):
        """Render this view on the screen"""
        self.fill_remaining_space()
        self.cut_info()
        self.minimize_info()
        self.display_line(self.y, self.x, self.info, Color.HOLIDAYS)


class DeadlineView(EventView):
    """Display a line with deadline icon and task name"""

    def __init__(self, stdscr, y, x, event, screen):
        super().__init__(stdscr, y, x, event, screen)
        self.info = f"{self.icon} {self.event.name[self.indent:]}"

    @property
    def icon(self):
        """Set the icon for task deadline"""
        return cf.DEADLINE_ICON

    @property
    def indent(self):
        """Calculate the left indentation depending on the task level"""
        if self.event.name[:4] == '----':
            return 4
        if self.event.name[:2] == '--':
            return 2
        return 0

    def render(self):
        """Render this view on the screen"""
        if self.screen.privacy or self.event.privacy:
            self.obfuscate_info()
        self.fill_remaining_space()
        self.cut_info()
        self.minimize_info()
        self.display_line(self.y, self.x, self.info, Color.DEADLINES)


class DailyView(View):
    """Display all events occurring on this days"""

    def __init__(self, stdscr, y, x, repeated_user_events, repeated_ics_events, user_events,
                 user_ics_events, holidays, birthdays, user_tasks, user_ics_tasks, screen,
                 index_offset, is_selection_day=True):

        super().__init__(stdscr, y, x)
        self.repeated_user_events = repeated_user_events.filter_events_that_day(screen)
        self.repeated_ics_events = repeated_ics_events.filter_events_that_day(screen)
        self.user_events = user_events.filter_events_that_day(screen)
        self.user_ics_events = user_ics_events.filter_events_that_day(screen)
        self.user_ics_events.items = sorted(self.user_ics_events.items, key=lambda x: (x.hour is None, x.hour))
        self.holidays = holidays.filter_events_that_day(screen)
        self.birthdays = birthdays.filter_events_that_day(screen)
        self.deadlines = user_tasks.filter_events_that_day(screen)
        self.deadlines_ics = user_ics_tasks.filter_events_that_day(screen)
        self.screen = screen
        self.index_offset = index_offset
        self.y_cell = (self.screen.y_max - 3) // 6
        self.x_cell = self.screen.x_max // 7
        self.hidden_events_sign = cf.HIDDEN_ICON + " "*(self.screen.x_max-self.x-len(cf.HIDDEN_ICON))
        self.num_events_this_day = 0
        self.is_selection_day = is_selection_day

    def render(self):
        """Render this view on the screen"""
        index = 0

        # Show user events:
        for event in self.user_events.items:
            if index >= self.y_cell - 1 and self.screen.calendar_state == CalState.MONTHLY:
                self.display_line(self.y + self.y_cell - 2, self.x, self.hidden_events_sign, Color.EVENTS)
            else:
                user_event_view = UserEventView(self.stdscr, self.y + index, self.x, event, self.screen)
                user_event_view.render()

                if self.screen.selection_mode and self.is_selection_day and self.screen.state == AppState.CALENDAR:
                    self.display_line(self.y + index, self.x, str(index + self.index_offset + 1), Color.ACTIVE_PANE)
            index += 1

        # Show repeated user events and events from ics:
        for event_list in [self.repeated_user_events.items, self.user_ics_events.items, self.repeated_ics_events.items]:
            for event in event_list:
                if index >= self.y_cell - 1 and self.screen.calendar_state == CalState.MONTHLY:
                    self.display_line(self.y + self.y_cell - 2, self.x, self.hidden_events_sign, Color.EVENTS)
                else:
                    user_event_view = UserEventView(self.stdscr, self.y + index, self.x, event, self.screen)
                    user_event_view.render()
                index += 1

        # Show deadlines for tasks, both from csv and ics files:
        for deadline_list in [self.deadlines.items, self.deadlines_ics.items]:
            for event in deadline_list:
                if index >= self.y_cell - 1 and self.screen.calendar_state == CalState.MONTHLY:
                    self.display_line(self.y + self.y_cell - 2, self.x, self.hidden_events_sign, Color.DEADLINES)
                else:
                    deadline_view = DeadlineView(self.stdscr, self.y + index, self.x, event, self.screen)
                    deadline_view.render()
                index += 1

        # Show holidays:
        if cf.DISPLAY_HOLIDAYS:
            for event in self.holidays.items:
                if index >= self.y_cell - 1 and self.screen.calendar_state == CalState.MONTHLY:
                    self.display_line(self.y + self.y_cell - 2, self.x, self.hidden_events_sign, Color.HOLIDAYS)
                else:
                    holiday_view = HolidayView(self.stdscr, self.y + index, self.x, event, self.screen)
                    holiday_view.render()
                index += 1

        # Show birthdays:
        if cf.BIRTHDAYS_FROM_ABOOK:
            for event in self.birthdays.items:
                if index >= self.y_cell - 1 and self.screen.calendar_state == CalState.MONTHLY:
                    self.display_line(self.y + self.y_cell - 2, self.x, self.hidden_events_sign, Color.BIRTHDAYS)
                else:
                    birthday_view = BirthdayView(self.stdscr, self.y + index, self.x, event, self.screen)
                    birthday_view.render()
                index += 1

        self.num_events_this_day = index


class DayNumberView(View):
    """Display the date of the day in month with proper styling"""

    def __init__(self, stdscr, y, x, screen, day, day_in_week, x_cell):
        super().__init__(stdscr, y, x)
        self.screen = screen
        self.day = day
        self.day_in_week = day_in_week
        self.x_cell = x_cell
        self.screen.day = self.day

    def render(self):
        """Render this view on the screen"""
        if self.screen.date == self.screen.today:
            today = f"{self.day}{cf.TODAY_ICON}{' '*(self.x_cell - len(str(self.day)) - 2)}"
            self.display_line(self.y, self.x, today, Color.TODAY, cf.BOLD_TODAY, cf.UNDERLINED_TODAY)
        elif self.day_in_week + 1 in cf.WEEKEND_DAYS:
            weekend = f"{self.day}{' '*(self.x_cell - len(str(self.day)) - 1)}"
            self.display_line(self.y, self.x, weekend, Color.WEEKENDS, cf.BOLD_WEEKENDS, cf.UNDERLINED_WEEKENDS)
        else:
            weekday = f"{self.day}{' '*(self.x_cell - len(str(self.day)) - 1)}"
            self.display_line(self.y, self.x, weekday, Color.DAYS, cf.BOLD_DAYS, cf.UNDERLINED_DAYS)


class TitleView(View):
    """Show the title in the header"""

    def __init__(self, stdscr, y, x, title, screen):
        super().__init__(stdscr, y, x)
        self.title = title
        self.screen = screen

    def render(self):
        """Render this view on the screen"""
        if self.screen.is_active_pane and self.screen.split:
            self.display_line(0, self.screen.x_min, self.title, Color.ACTIVE_PANE, cf.BOLD_ACTIVE_PANE, cf.UNDERLINED_ACTIVE_PANE)
        else:
            self.display_line(0, self.screen.x_min, self.title, Color.CALENDAR_HEADER, cf.BOLD_TITLE, cf.UNDERLINED_TITLE)


class HeaderView(View):
    """Show the header that includes the weather, time, and title"""

    def __init__(self, stdscr, y, x, title, weather, screen):
        super().__init__(stdscr, y, x)
        self.title = title
        self.weather = weather
        self.screen = screen

    def render(self):
        """Render this view on the screen"""
        _, x_max = self.stdscr.getmaxyx()

        # Show title:
        title_view = TitleView(self.stdscr, 0, self.screen.x_min, self.title, self.screen)
        title_view.render()

        if self.screen.currently_drawn == AppState.JOURNAL and self.screen.split:
            return

        # Show weather is space allows and it is loaded:
        size_allows = len(self.weather.forcast) < self.screen.x_max - len(self.title)
        if cf.SHOW_WEATHER and size_allows:
            self.display_line(0, self.screen.x_max - len(self.weather.forcast) - 1, self.weather.forcast, Color.WEATHER)

        # Show time:
        time_string = time.strftime("%H:%M", time.localtime())
        size_allows = len(self.weather.forcast) < self.screen.x_max - len(self.title) - len(time_string)
        if cf.SHOW_CURRENT_TIME and size_allows:
            self.display_line(0, (self.screen.x_max // 2 - 2), time_string, Color.TIME)


class FooterView(View):
    """Display the footer with keybinding"""

    def __init__(self, stdscr, y, x, screen):
        super().__init__(stdscr, y, x)
        self.screen = screen

    def render(self):
        """Render this view on the screen"""
        if not cf.SHOW_KEYBINDINGS: return
        clear_line(self.stdscr, self.screen.y_max - 1)
        if self.screen.state == AppState.CALENDAR:
            if self.screen.calendar_state == CalState.MONTHLY:
                hint = CALENDAR_HINT
            else:
                hint = CALENDAR_HINT_D
        elif self.screen.state == AppState.JOURNAL:
            hint = JOURNAL_HINT
        self.display_line(self.screen.y_max - 1, 0, hint, Color.HINTS)


class ErrorView(View):
    """Display the error messages"""

    def __init__(self, stdscr, y, x, screen):
        super().__init__(stdscr, y, x)
        self.screen = screen
        self.error = error


    def render(self):
        """Render this view on the screen"""
        if self.error.has_occured:
            clear_line(self.stdscr, self.screen.y_max - 2)

            # Depending on error type, display different messages:
            if self.error.number_of_errors > 1 or "ERROR" in self.error.type:
                self.display_line(self.screen.y_max - 2, 0, MSG_ERRORS, Color.IMPORTANT)
            else:
                self.display_line(self.screen.y_max - 2, 0, MSG_INPUT, Color.HINTS)

            self.error.clear_buffer()


class SeparatorView(View):
    """Display the separator in the split screen"""

    def __init__(self, stdscr, y, x, screen):
        super().__init__(stdscr, y, x)
        self.screen = screen

    def render(self):
        """Render this view on the screen"""
        _, x_max = self.stdscr.getmaxyx()
        x_separator = x_max - self.screen.journal_pane_width
        y_cell = (self.screen.y_max - 3) // 6
        height = self.screen.number_of_weeks * y_cell + 2 if cf.SHOW_CALENDAR_BORDERS else self.screen.y_max
        for row in range(height):
            self.display_line(row, x_separator, cf.SEPARATOR_ICON, Color.SEPARATOR)

        if cf.SHOW_CALENDAR_BORDERS and self.screen.calendar_state == CalState.MONTHLY:
            for row in range(1, self.screen.number_of_weeks):
                self.display_line(row * y_cell + 1, x_separator, "┤", Color.CALENDAR_BORDER)
            self.display_line(self.screen.number_of_weeks * y_cell + 1, x_separator, "┘", Color.CALENDAR_BORDER)


class CalenarBorderView(View):
    """Display the borders in the monthly view"""

    def __init__(self, stdscr, y, x, screen):
        super().__init__(stdscr, y, x)
        self.screen = screen
        self.rows = self.screen.number_of_weeks

    def render(self):
        """Render this view on the screen"""
        x_cell = self.screen.x_max // 7
        y_cell = (self.screen.y_max - 3) // 6

        # Vertical lines:
        for column in range(1, 7):
            for row in range(y_cell * self.rows):
                self.display_line(row + 2, column * x_cell - 1, "│", Color.CALENDAR_BORDER)

        # Horizontal lines:
        for row in range(1, self.rows + 1):
            for column in range(0, self.screen.x_max):
                self.display_line(row * y_cell + 1, column, "─", Color.CALENDAR_BORDER)

        # Connectors:
        for row in range(1, self.rows):
            for column in range(1, 7):
                self.display_line(row * y_cell + 1, column*x_cell - 1, "┼", Color.CALENDAR_BORDER)
        for column in range(1, 7):
            self.display_line(self.rows * y_cell + 1, column*x_cell - 1, "┴", Color.CALENDAR_BORDER)


class DaysNameView(View):
    """Display day name depending on the screen available and with right style"""

    def __init__(self, stdscr, y, x, screen):
        super().__init__(stdscr, y, x)
        self.screen = screen

    def render(self):
        """Render this view on the screen"""
        num = 2 if self.screen.x_max < 74 else 12
        x_cell = int(self.screen.x_max // 7)

        # Depending on which day we start the week, weekends are shifted:
        for i in range(7):
            shift = cf.START_WEEK_DAY - 1
            day_number = i + shift - 7*((i + shift) > 6)
            day_names = DAYS_PERSIAN if cf.USE_PERSIAN_CALENDAR else DAYS
            name = day_names[day_number][:num]
            x = self.x + i*x_cell
            if day_number + 1 not in cf.WEEKEND_DAYS:
                self.display_line(self.y, x, name, Color.DAY_NAMES, cf.BOLD_DAY_NAMES, cf.UNDERLINED_DAY_NAMES)
            else:
                self.display_line(self.y, x, name, Color.WEEKEND_NAMES, cf.BOLD_WEEKEND_NAMES, cf.UNDERLINED_WEEKEND_NAMES)


##################### SCREENS ##########################


class DailyScreenView(View):
    """Daily view showing events of the day"""

    def __init__(self, stdscr, y, x, weather, user_events, user_ics_events, holidays, birthdays, user_tasks, user_ics_tasks, screen):
        super().__init__(stdscr, y, x)
        self.weather = weather
        self.user_events = user_events
        self.user_ics_events = user_ics_events
        self.holidays = holidays
        self.birthdays = birthdays
        self.user_tasks = user_tasks
        self.user_ics_tasks = user_ics_tasks
        self.screen = screen

    @property
    def dates(self):
        """Return dates that exist in this month"""
        return Calendar(cf.START_WEEK_DAY - 1, cf.USE_PERSIAN_CALENDAR).monthdayscalendar(self.screen.year, self.screen.month)

    @property
    def week_day(self):
        """Number of the day in week, from 0 to 6"""
        for week in self.dates:
            if self.screen.day not in week:
                continue
            weekday = (week.index(self.screen.day) + (cf.START_WEEK_DAY - 1)) % 7
        return weekday

    @property
    def color(self):
        """Color of the date and dayname"""
        if self.screen.date == self.screen.today:
            return Color.TODAY
        if self.week_day + 1 in cf.WEEKEND_DAYS:
            return Color.WEEKENDS
        return Color.DAYS

    @property
    def icon(self):
        """Icon of today"""
        return cf.TODAY_ICON if self.screen.date == self.screen.today else ''

    def render(self):
        """Render this view on the screen"""
        self.screen.currently_drawn = AppState.CALENDAR
        if self.screen.x_max < 6 or self.screen.y_max < 3:
            return

        # Form a string with month, year, and day with today icon:
        month_names = MONTHS_PERSIAN if cf.USE_PERSIAN_CALENDAR else MONTHS
        month_string = str(month_names[self.screen.month-1])
        date_string = f'{month_string} {self.screen.year}'

        # Display header and footer:
        header_view = HeaderView(self.stdscr, 0, 0, date_string, self.weather, self.screen)
        header_view.render()

        # Display the events from current day to as many as possible days:
        repeated_user_events = RepeatedEvents(self.user_events, cf.USE_PERSIAN_CALENDAR, self.screen.year)
        repeated_ics_events = RepeatedEvents(self.user_ics_events, cf.USE_PERSIAN_CALENDAR, self.screen.year)
        max_num_days = (self.screen.y_max - 5)//2
        vertical_shift = 0

        is_selection_day = True
        for _ in range(max_num_days):

            # Display day name:
            day_string = f"{self.screen.day} {DAYS[self.week_day]} {self.icon}"
            day_string += " " * (max([len(day_name) for day_name in DAYS]) + 3 + len(self.icon) - len(day_string))
            self.display_line(self.y + 2 + vertical_shift, self.x, day_string, self.color)

            # Display events of the day:
            daily_view = DailyView(self.stdscr, self.y + 3 + vertical_shift, self.x, repeated_user_events,
                                   repeated_ics_events, self.user_events, self.user_ics_events, self.holidays,
                                   self.birthdays, self.user_tasks, self.user_ics_tasks, self.screen, 0,
                                   is_selection_day)
            daily_view.render()

            # Move to the next day:
            vertical_shift += daily_view.num_events_this_day + 2
            self.screen.next_day()
            is_selection_day = False

        # Return screen day back to the original day:
        for _ in range(max_num_days):
            self.screen.previous_day()


class MonthlyScreenView(View):
    """Monthly view showing events of the month"""

    def __init__(self, stdscr, y, x, weather, user_events, user_ics_events, holidays, birthdays, user_tasks, user_ics_tasks, screen):
        super().__init__(stdscr, y, x)
        self.weather = weather
        self.user_events = user_events
        self.user_ics_events = user_ics_events
        self.holidays = holidays
        self.birthdays = birthdays
        self.user_tasks = user_tasks
        self.user_ics_tasks = user_ics_tasks
        self.screen = screen

    def render(self):
        """Render this view on the screen"""
        self.screen.currently_drawn = AppState.CALENDAR
        if self.screen.x_max < 6 or self.screen.y_max < 3: return

        # Info about the month:
        month_names = MONTHS_PERSIAN if cf.USE_PERSIAN_CALENDAR else MONTHS
        month_year_string = month_names[self.screen.month-1] + " " + str(self.screen.year)
        dates = Calendar(cf.START_WEEK_DAY - 1, cf.USE_PERSIAN_CALENDAR).monthdayscalendar(self.screen.year, self.screen.month)

        y_cell = (self.screen.y_max - 3) // 6
        x_cell = self.screen.x_max // 7

        header_view = HeaderView(self.stdscr, 0, 0, month_year_string, self.weather, self.screen)
        days_name_view = DaysNameView(self.stdscr, 1, 0, self.screen)
        header_view.render()
        days_name_view.render()

        # Displaying the dates and events:
        repeated_user_events = RepeatedEvents(self.user_events, cf.USE_PERSIAN_CALENDAR, self.screen.year)
        repeated_ics_events = RepeatedEvents(self.user_ics_events, cf.USE_PERSIAN_CALENDAR, self.screen.year)
        num_events_this_month = 0
        for row, week in enumerate(dates):
            for col, day in enumerate(week):
                if day != 0:
                    # Display dates of the month with proper styles:
                    day_in_week = col + (cf.START_WEEK_DAY - 1) - 7 * ((col + (cf.START_WEEK_DAY - 1)) > 6)
                    day_number_view = DayNumberView(self.stdscr, 2 + row * y_cell, col * x_cell, self.screen, day, day_in_week, x_cell)
                    day_number_view.render()

                    # Display the events:
                    self.screen.day = day
                    daily_view = DailyView(self.stdscr, 3 + row * y_cell, col * x_cell, repeated_user_events,
                                           repeated_ics_events, self.user_events, self.user_ics_events,
                                           self.holidays, self.birthdays, self.user_tasks, self.user_ics_tasks,
                                           self.screen, num_events_this_month)
                    daily_view.render()
                    num_events_this_month += len(self.user_events.filter_events_that_day(self.screen).items)

        if cf.SHOW_CALENDAR_BORDERS:
            calendar_border_view = CalenarBorderView(self.stdscr, 0, 0, self.screen)
            calendar_border_view.render()


class JournalScreenView(View):
    def __init__(self, stdscr, y, x, weather, user_tasks, user_ics_tasks, screen):
        super().__init__(stdscr, y, x)
        self.weather = weather
        self.user_tasks = user_tasks
        self.user_ics_tasks = user_ics_tasks
        self.screen = screen

    def render(self):
        """Journal view showing all tasks"""
        self.screen.currently_drawn = AppState.JOURNAL
        if self.screen.x_max < 6 or self.screen.y_max < 3:
            return

        # Display header and footer:
        header_view = HeaderView(self.stdscr, 0, 0, cf.JOURNAL_HEADER, self.weather, self.screen)
        header_view.render()

        # Display the tasks:
        journal_view = JournalView(self.stdscr, 2, self.screen.x_min, self.user_tasks, self.user_ics_tasks, self.screen)
        journal_view.render()


class WelcomeScreenView(View):
    """Welcome screen displaying greeting info on the first run"""

    def __init__(self, stdscr, y, x, screen):
        super().__init__(stdscr, y, x)
        self.screen = screen

    def calibrate_position(self):
        """Depending on the screen space calculate the best position"""
        self.y_max, self.x_max = self.stdscr.getmaxyx()

    def render(self):
        """Draw the welcome screen"""
        self.calibrate_position()
        self.stdscr.clear()
        self.fill_background()

        if self.x_max < len(MSG_WELCOME_4)+2 or self.y_max < 12:
            self.display_line(0, 0, "Welcome!", Color.ACTIVE_PANE)
            return

        d_x = self.x_max//2
        d_y = self.y_max//2 - 5

        self.display_line(d_y, d_x - len(MSG_WELCOME_1+__version__+" ")//2, f"{MSG_WELCOME_1} {__version__}", Color.ACTIVE_PANE)
        self.display_line(d_y + 1, d_x - len(MSG_WELCOME_2)//2, MSG_WELCOME_2, Color.TODO)
        self.display_line(d_y + 3, d_x - len(MSG_WELCOME_3)//2, MSG_WELCOME_3, Color.TODO)
        self.display_line(d_y + 4, d_x - len(str(cf.config_folder))//2, str(cf.config_folder), Color.TITLE)
        self.display_line(d_y + 6, d_x - len(MSG_WELCOME_4)//2, MSG_WELCOME_4, Color.TODO)
        self.display_line(d_y + 7, d_x - len(MSG_SITE)//2, MSG_SITE, Color.TITLE)
        self.display_line(d_y + 9, d_x - len(MSG_WELCOME_5)//2, MSG_WELCOME_5, Color.TODO)


class HelpScreenView(View):
    """Help screen displaying information about keybindings"""

    def __init__(self, stdscr, y, x, screen):
        super().__init__(stdscr, y, x)
        self.screen = screen

    def calibrate_position(self):
        """Depending on the screen space calculate the best position"""
        self.y_max, self.x_max = self.stdscr.getmaxyx()

        if self.x_max < 102:
            self.global_shift_x = 0
            self.shift_x = 0
            self.shift_y = 6 + len(KEYS_GENERAL) + len(KEYS_CALENDAR)
        else:
            self.global_shift_x = (self.x_max - 102) // 2
            self.shift_x = 45
            self.shift_y = 2

        if self.y_max > 20 and self.x_max >= 102:
            self.global_shift_y = (self.y_max - 25) // 2
        else:
            self.global_shift_y = 0

    def render(self):
        """Draw the help screen"""
        self.calibrate_position()
        if self.x_max < 6 or self.y_max < 3:
            return
        self.stdscr.clear()
        self.fill_background()

        # Left column:
        self.display_line(self.global_shift_y, self.global_shift_x + 1, f"{MSG_NAME} {__version__}",
                            Color.ACTIVE_PANE, cf.BOLD_TITLE, cf.UNDERLINED_TITLE)
        self.display_line(self.global_shift_y + 2, self.global_shift_x + 8,
                          TITLE_KEYS_GENERAL, Color.TITLE, cf.BOLD_TITLE, cf.UNDERLINED_TITLE)
        for index, key in enumerate(KEYS_GENERAL):
            self.display_line(self.global_shift_y + index + 3, self.global_shift_x, key, Color.ACTIVE_PANE)
            self.display_line(self.global_shift_y + index + 3, self.global_shift_x + 8, KEYS_GENERAL[key], Color.TODO)

        self.display_line(self.global_shift_y + 4 + len(KEYS_GENERAL), self.global_shift_x + 8,
                          TITLE_KEYS_CALENDAR, Color.TITLE, cf.BOLD_TITLE, cf.UNDERLINED_TITLE)
        for index, key in enumerate(KEYS_CALENDAR):
            self.display_line(self.global_shift_y + index + 5 + len(KEYS_GENERAL), self.global_shift_x, key, Color.ACTIVE_PANE)
            self.display_line(self.global_shift_y + index + 5 + len(KEYS_GENERAL), self.global_shift_x + 8,
                                                                            KEYS_CALENDAR[key], Color.TODO)

        # Right column:
        d_x = self.global_shift_x + self.shift_x
        d_y = self.global_shift_y + self.shift_y
        self.display_line(d_y, d_x + 8, TITLE_KEYS_JOURNAL, Color.TITLE, cf.BOLD_TITLE, cf.UNDERLINED_TITLE)
        for index, key in enumerate(KEYS_TODO):
            self.display_line(d_y + index + 1, d_x, key, Color.ACTIVE_PANE)
            self.display_line(d_y + index + 1, d_x + 8, KEYS_TODO[key], Color.TODO)

        # Additional info:
        d_x = self.global_shift_x + self.shift_x + 8
        d_y = self.global_shift_y + len(KEYS_TODO) + self.shift_y
        self.display_line(d_y + 2, d_x, MSG_VIM, Color.ACTIVE_PANE)
        self.display_line(d_y + 4, d_x, MSG_INFO, Color.TODO)
        self.display_line(d_y + 5, d_x, MSG_SITE, Color.TITLE)


def main(stdscr) -> None:
    """Main function that runs and switches screens"""

    # Load the weather:
    weather = Weather(cf.WEATHER_CITY, cf.WEATHER_METRIC_UNITS)
    if cf.SHOW_WEATHER:
        threading.Thread(target=weather.load_from_wttr).start()

    screen = Screen(stdscr, cf)

    # Initialise loaders:
    event_loader_csv = EventLoaderCSV(cf)
    task_loader_csv = TaskLoaderCSV(cf)
    event_loader_ics = EventLoaderICS(cf)
    task_loader_ics = TaskLoaderICS(cf)
    birthday_loader = BirthdayLoader(cf)
    holiday_loader = HolidayLoader(cf)

    # Load the data:
    user_events = event_loader_csv.load()
    user_tasks = task_loader_csv.load()
    user_ics_events = event_loader_ics.load()
    user_ics_tasks = task_loader_ics.load()
    holidays = holiday_loader.load()
    birthdays = birthday_loader.load()

    # Initialise savers and importers:
    event_saver_csv = EventSaverCSV(user_events, cf)
    task_saver_csv = TaskSaverCSV(user_tasks, cf)
    importer = Importer(user_tasks, user_events, cf)

    read_items_from_user_arguments(screen, user_tasks, user_events, task_saver_csv, event_saver_csv)

    # Initialise terminal screen:
    stdscr = curses.initscr()
    curses.noecho()
    curses.curs_set(False)
    initialize_colors(cf)

    # Initialise screen views:
    app_view = View(stdscr, 0, 0)
    monthly_screen_view = MonthlyScreenView(stdscr, 0, 0, weather, user_events, user_ics_events,
                                            holidays, birthdays, user_tasks, user_ics_tasks, screen)
    daily_screen_view = DailyScreenView(stdscr, 0, 0, weather, user_events, user_ics_events,
                                        holidays, birthdays, user_tasks, user_ics_tasks, screen)
    journal_screen_view = JournalScreenView(stdscr, 0, 0, weather, user_tasks, user_ics_tasks, screen)
    help_screen_view = HelpScreenView(stdscr, 0, 0, screen)
    welcome_screen_view = WelcomeScreenView(stdscr, 0, 0, screen)
    footer_view = FooterView(stdscr, 0, 0, screen)
    separator_view = SeparatorView(stdscr, 0, 0, screen)
    error_view = ErrorView(stdscr, 0, 0, screen)

    # Show welcome screen on the first run:
    if cf.is_first_run:
        screen.state = AppState.WELCOME
    while screen.state == AppState.WELCOME:
        welcome_screen_view.render()
        control_welcome_screen(stdscr, screen)

    # Running different screens depending on the state:
    while screen.state != AppState.EXIT:
        stdscr.clear()
        app_view.fill_background()

        # Calculate screen refreh rate:
        curses.halfdelay(200)
        if user_tasks.has_active_timer and screen.state == AppState.JOURNAL:
            curses.halfdelay(cf.REFRESH_INTERVAL * 10)

        # Calendar screens:
        if screen.state == AppState.CALENDAR:

            if screen.calendar_state == CalState.MONTHLY:
                monthly_screen_view.render()
            else:
                daily_screen_view.render()

            if screen.split:
                journal_screen_view.render()
                separator_view.render()
            footer_view.render()
            error_view.render()

            if screen.calendar_state == CalState.MONTHLY:
                control_monthly_screen(stdscr, screen, user_events, importer)
            else:
                control_daily_screen(stdscr, screen, user_events, importer)

        # Journal screen:
        elif screen.state == AppState.JOURNAL:
            if screen.split:
                if screen.calendar_state == CalState.MONTHLY:
                    monthly_screen_view.render()
                else:
                    daily_screen_view.render()
                separator_view.render()
            journal_screen_view.render()
            footer_view.render()
            error_view.render()
            control_journal_screen(stdscr, screen, user_tasks, importer)

        # Help screen:
        elif screen.state == AppState.HELP:
            help_screen_view.render()
            control_help_screen(stdscr, screen)

        else:
            break

        # If something has been changed, save the data:
        if user_events.changed:
            event_saver_csv.save()
            screen.refresh_now = True
        if user_tasks.changed:
            task_saver_csv.save()
            screen.refresh_now = True

        # If needed, reload the data:
        if screen.is_time_to_reload:
            user_events = event_loader_csv.load()
            user_tasks = task_loader_csv.load()
            user_ics_events = event_loader_ics.load()
            user_ics_tasks = task_loader_ics.load()

    # Cleaning up before quitting:
    curses.echo()
    curses.curs_set(True)
    curses.endwin()


def cli() -> None:
    try:
        curses.wrapper(main)
    except (KeyboardInterrupt, curses.error): # Hides strange curses quitting error
        pass


if __name__ == "__main__":
    cli()



""" Module that controls interactions with the user, like questions and confirmations"""

import curses
import logging

from calcure.data import Frequency
from calcure.colors import Color



def safe_run(func):
    """Decorator preventing crashes on keyboard interruption and no input"""

    def inner(stdscr, screen, *args, **kwargs):
        try:
            func(stdscr, screen, *args, **kwargs)

        # Handle keyboard interruption with ctrl+c:
        except KeyboardInterrupt:
            pass

        # Prevent crash if no input:
        except curses.error:
            pass
    return inner


def display_question(stdscr, y, x, question, color):
    """Display the line of text respecting the styling and available space"""
    y_max, x_max = stdscr.getmaxyx()
    if y >= y_max or x >= x_max:
        return
    question = question[:(x_max - x - 1)]
    stdscr.addstr(y, x, question, curses.color_pair(color.value))


def clear_line(stdscr, y, x=0):
    """Clear a line from any text"""
    _, x_max = stdscr.getmaxyx()
    stdscr.addstr(y, x, " " * (x_max - x - 1), curses.color_pair(Color.EMPTY.value))


def input_string(stdscr, y, x, question, answer_length):
    """Ask user to input something and return it as a string"""
    curses.echo()
    curses.curs_set(True)
    display_question(stdscr, y, x, question, Color.PROMPTS)
    stdscr.refresh()
    try:
        answer = stdscr.getstr(y, len(question) + x, answer_length).decode(encoding="utf-8")
    except (UnicodeDecodeError, KeyboardInterrupt):
        answer = ""
        logging.warning("Incorrect characters in user input.")
    curses.noecho()
    curses.curs_set(False)
    return answer


def input_integer(stdscr, y, x, question):
    """Ask user for an integer number and check if it is an integer"""
    number = input_string(stdscr, y, x, question, 3)
    try:
        number = int(number) - 1
    except (ValueError, KeyboardInterrupt):
        logging.warning("Incorrect number input.")
        return None
    return number


def input_day(stdscr, y, x, prompt_string):
    """Ask user for an integer number and check if it is an integer"""
    number = input_string(stdscr, y, x, prompt_string, 2)
    try:
        number = int(number)
    except (ValueError, KeyboardInterrupt):
        logging.warning("Incorrect day input.")
        return None
    return number


def input_date(stdscr, y, x, prompt_string):
    """Ask user to input date in YYYY/MM/DD format and check if it was a valid entry"""
    date_unformated = input_string(stdscr, y, x, prompt_string, 10)
    try:
        year = int(date_unformated.split("/")[0])
        month = int(date_unformated.split("/")[1])
        day = int(date_unformated.split("/")[2])
        return year, month, day
    except (ValueError, IndexError, KeyboardInterrupt):
        logging.warning("Incorrect date input.")
        return None, None, None


def input_frequency(stdscr, y, x, question):
    """Ask user for the frequency of event repetitions"""
    freq = input_string(stdscr, y, x, question, 2)
    if freq == 'd':
        return Frequency.DAILY
    if freq == 'w':
        return Frequency.WEEKLY
    if freq == 'm':
        return Frequency.MONTHLY
    if freq == 'y':
        return Frequency.YEARLY
    if freq == 'n':
        return Frequency.ONCE
    return None


def ask_confirmation(stdscr, question, confirmations_enabled):
    """Ask user confirmation for an action"""
    if not confirmations_enabled:
        return True
    y_max, _ = stdscr.getmaxyx()
    curses.halfdelay(255)
    display_question(stdscr, y_max - 2, 0, question, Color.CONFIRMATIONS)
    key = stdscr.getkey()
    confirmed = (key == "y")
    return confirmed


def vim_style_exit(stdscr, screen):
    """Handle vim style key combinations like ZZ and ZQ for exit"""
    if screen.key == "Z":
        try:
            screen.key = stdscr.getkey()
            return (screen.key in ["Z", "Q"])
        except KeyboardInterrupt:
            return True
    else:
        return False



"""This module contains functions that react to user input on each screen"""

import curses
import importlib

# Modules:
from calcure.data import *
from calcure.dialogues import *
from calcure.configuration import Config

cf = Config()

# Language:
if cf.LANG == "fr":
    from calcure.translations.fr import *
elif cf.LANG == "ru":
    from calcure.translations.ru import *
elif cf.LANG == "it":
    from calcure.translations.it import *
elif cf.LANG == "br":
    from calcure.translations.br import *
elif cf.LANG == "tr":
    from calcure.translations.tr import *
elif cf.LANG == "zh":
    from calcure.translations.zh import *
elif cf.LANG == "sk":
    from calcure.translations.sk import *
else:
    from calcure.translations.en import *


def safe_run(func):
    """Decorator preventing crashes on keyboard interruption and no input"""

    def inner(stdscr, screen, *args, **kwargs):
        try:
            func(stdscr, screen, *args, **kwargs)

        # Handle keyboard interruption with ctrl+c:
        except KeyboardInterrupt:
            confirmed = ask_confirmation(stdscr, MSG_EXIT, cf.ASK_CONFIRMATION_TO_QUIT)
            screen.state = AppState.EXIT if confirmed else screen.state

        # Prevent crash if no input:
        except curses.error:
            pass
    return inner


@safe_run
def control_monthly_screen(stdscr, screen, user_events, importer):
    """Handle user input on the montly screen"""

    # If we previously entered the selection mode, now we perform the action:
    if screen.selection_mode:
        screen.selection_mode = False

        # Change event status:
        if screen.key in ['i', 'h']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_HIGH)
            if user_events.filter_events_that_month(screen).is_valid_number(number):
                event_id = user_events.filter_events_that_month(screen).items[number].item_id
                user_events.toggle_item_status(event_id, Status.IMPORTANT)
        if screen.key == 'l':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_LOW)
            if user_events.filter_events_that_month(screen).is_valid_number(number):
                event_id = user_events.filter_events_that_month(screen).items[number].item_id
                user_events.toggle_item_status(event_id, Status.UNIMPORTANT)
        if screen.key == 'u':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_RESET)
            if user_events.filter_events_that_month(screen).is_valid_number(number):
                event_id = user_events.filter_events_that_month(screen).items[number].item_id
                user_events.toggle_item_status(event_id, Status.NORMAL)
        if screen.key == 'd':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_DONE)
            if user_events.filter_events_that_month(screen).is_valid_number(number):
                event_id = user_events.filter_events_that_month(screen).items[number].item_id
                user_events.toggle_item_status(event_id, Status.DONE)

        # Toggle event privacy:
        if screen.key == '.':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_PRIVACY)
            if user_events.filter_events_that_month(screen).is_valid_number(number):
                event_id = user_events.filter_events_that_month(screen).items[number].item_id
                user_events.toggle_item_privacy(event_id)

        # Delete event:
        if screen.key == 'x':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_DEL)
            if user_events.filter_events_that_month(screen).is_valid_number(number):
                event_id = user_events.filter_events_that_month(screen).items[number].item_id
                user_events.delete_item(event_id)

        # Rename event:
        if screen.key in ['e', 'r']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_REN)
            if user_events.filter_events_that_month(screen).is_valid_number(number):
                event_id = user_events.filter_events_that_month(screen).items[number].item_id
                clear_line(stdscr, screen.y_max-2)
                new_name = input_string(stdscr, screen.y_max-2, 0, MSG_NEW_TITLE, screen.x_max-len(MSG_NEW_TITLE)-2)
                user_events.rename_item(event_id, new_name)

        # Move event:
        if screen.key in ['m', 'M']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_MV)
            if user_events.filter_events_that_month(screen).is_valid_number(number):
                event_id = user_events.filter_events_that_month(screen).items[number].item_id
                clear_line(stdscr, screen.y_max-2)
                if screen.key == 'm':
                    year, month, day = input_date(stdscr, screen.y_max-2, 0, MSG_EVENT_MV_TO)
                    if screen.is_valid_date(year, month, day):
                        user_events.change_date(event_id, year, month, day)

                if screen.key == 'M':
                    question = f'{MSG_EVENT_MV_TO_D} {screen.year}/{screen.month}/'
                    day = input_day(stdscr, screen.y_max-2, 0, question)
                    if screen.is_valid_day(day):
                        user_events.change_day(event_id, day)

    # Otherwise, we check for user input:
    else:
        # Wait for user to press a key:
        screen.key = stdscr.getkey()

        # If we need to select an event, change to selection mode:
        selection_keys = ['h', 'l', 'u', 'i', 'd', 'x', 'e', 'r', 'c', 'm', 'M', '.']
        if screen.key in selection_keys and user_events.filter_events_that_month(screen).items:
            screen.selection_mode = True

        # Navigation:
        if screen.key in ["n", "j", "KEY_UP", "KEY_RIGHT"]:
            screen.next_month()
        if screen.key in ["p", "k", "KEY_DOWN", "KEY_LEFT"]:
            screen.previous_month()
        if screen.key in ["KEY_HOME", "R"]:
            screen.reset_to_today()

        # Handle "g" and "G" as go to selected day:
        if screen.key == "g":
            clear_line(stdscr, screen.y_max-2, 0)
            year, month, day = input_date(stdscr, screen.y_max-2, 0, MSG_GOTO)
            if screen.is_valid_date(year, month, day):
                screen.day = day
                screen.month = month
                screen.year = year
                screen.calendar_state = CalState.DAILY

        if screen.key == "G":
            clear_line(stdscr, screen.y_max-2, 0)
            question = f"{MSG_GOTO_D} {screen.year}/{screen.month}/"
            day = input_day(stdscr, screen.y_max-2, 0, question)
            if screen.is_valid_date(screen.year, screen.month, day):
                screen.day = day
                screen.calendar_state = CalState.DAILY

        # Change the view to daily:
        if screen.key == "v":
            screen.day = 1
            screen.calendar_state = CalState.DAILY

        # Add single event:
        if screen.key == "a":
            question = f'{MSG_EVENT_DATE} {screen.year}/{screen.month}/'
            day = input_day(stdscr, screen.y_max-2, 0, question)
            if screen.is_valid_day(day):
                clear_line(stdscr, screen.y_max-2)
                name = input_string(stdscr, screen.y_max-2, 0, MSG_EVENT_TITLE, screen.x_max-len(MSG_EVENT_TITLE)-2)
                event_id = user_events.items[-1].item_id + 1 if not user_events.is_empty() else 1
                user_events.add_item(UserEvent(event_id, screen.year, screen.month, day, name, 1, Frequency.ONCE, Status.NORMAL, False))

        # Add a recurring event:
        if screen.key == "A":
            question = f'{MSG_EVENT_DATE}{screen.year}/{screen.month}/'
            day = input_day(stdscr, screen.y_max-2, 0, question)
            if screen.is_valid_day(day):
                clear_line(stdscr, screen.y_max-2)
                name = input_string(stdscr, screen.y_max-2, 0, MSG_EVENT_TITLE, screen.x_max-len(MSG_EVENT_TITLE)-2)
                item_id = user_events.items[-1].item_id + 1 if not user_events.is_empty() else 1
                reps = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_REP)
                freq = input_frequency(stdscr, screen.y_max-2, 0, MSG_EVENT_FR)
                if reps is not None and freq is not None:
                    reps = 1 if reps == 0 else reps
                    user_events.add_item(UserEvent(item_id, screen.year, screen.month, day, name, reps+1, freq, Status.NORMAL, False))

        # Reload:
        if screen.key in ["Q"]:
            screen.reload_data = True
            screen.refresh_now = True

        # Imports:
        if screen.key == "C":
            confirmed = ask_confirmation(stdscr, MSG_EVENT_IMP, cf.ASK_CONFIRMATIONS)
            if confirmed:
                importer.import_events_from_calcurse()
                screen.refresh_now = True

        # Other actions:
        if vim_style_exit(stdscr, screen):
            confirmed = ask_confirmation(stdscr, MSG_EXIT, cf.ASK_CONFIRMATION_TO_QUIT)
            screen.state = AppState.EXIT if confirmed else screen.state
        if screen.key == "*":
            screen.privacy = not screen.privacy
        if screen.key in [" ", "KEY_BTAB"]:
            screen.state = AppState.JOURNAL
        if screen.key == "?":
            screen.state = AppState.HELP
        if screen.key == "q":
            confirmed = ask_confirmation(stdscr, MSG_EXIT, cf.ASK_CONFIRMATION_TO_QUIT)
            screen.state = AppState.EXIT if confirmed else screen.state
        if screen.key in ["/"]:
            screen.split = not screen.split
            screen.refresh_now = True


@safe_run
def control_daily_screen(stdscr, screen, user_events, importer):
    """Handle user input on the daily screen"""
    # If we previously entered the selection mode, now we perform the action:
    if screen.selection_mode:
        screen.selection_mode = False

        # Change event status:
        if screen.key in ['i', 'h']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_HIGH)
            if user_events.filter_events_that_day(screen).is_valid_number(number):
                item_id = user_events.filter_events_that_day(screen).items[number].item_id
                user_events.toggle_item_status(item_id, Status.IMPORTANT)
        if screen.key == 'l':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_LOW)
            if user_events.filter_events_that_day(screen).is_valid_number(number):
                item_id = user_events.filter_events_that_day(screen).items[number].item_id
                user_events.toggle_item_status(item_id, Status.UNIMPORTANT)
        if screen.key == 'u':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_RESET)
            if user_events.filter_events_that_day(screen).is_valid_number(number):
                item_id = user_events.filter_events_that_day(screen).items[number].item_id
                user_events.toggle_item_status(item_id, Status.NORMAL)
        if screen.key == 'd':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_DONE)
            if user_events.filter_events_that_day(screen).is_valid_number(number):
                item_id = user_events.filter_events_that_day(screen).items[number].item_id
                user_events.toggle_item_status(item_id, Status.DONE)

        # Toggle event privacy:
        if screen.key == '.':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_PRIVACY)
            if user_events.filter_events_that_day(screen).is_valid_number(number):
                event_id = user_events.filter_events_that_day(screen).items[number].item_id
                user_events.toggle_item_privacy(event_id)

        # Delete event:
        if screen.key in ['x']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_DEL)
            if user_events.filter_events_that_day(screen).is_valid_number(number):
                item_id = user_events.filter_events_that_day(screen).items[number].item_id
                user_events.delete_item(item_id)

        # Rename event:
        if screen.key in ['e', 'r']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_REN)
            if user_events.filter_events_that_day(screen).is_valid_number(number):
                item_id = user_events.filter_events_that_day(screen).items[number].item_id
                clear_line(stdscr, screen.y_max-2)
                new_name = input_string(stdscr, screen.y_max-2, 0, MSG_NEW_TITLE, screen.x_max-len(MSG_NEW_TITLE)-2)
                user_events.rename_item(item_id, new_name)

        # Move event:
        if screen.key in ['m', 'M']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_MV)
            if user_events.filter_events_that_day(screen).is_valid_number(number):
                item_id = user_events.filter_events_that_day(screen).items[number].item_id
                clear_line(stdscr, screen.y_max-2)
                if screen.key == 'm':
                    year, month, day = input_date(stdscr, screen.y_max-2, 0, MSG_EVENT_MV_TO)
                    if screen.is_valid_date(year, month, day):
                        user_events.change_date(event_id, year, month, day)

                if screen.key == 'M':
                    question = f'{MSG_EVENT_MV_TO_D}{screen.year}/{screen.month}/'
                    day = input_day(stdscr, screen.y_max-2, 0, question)
                    if screen.is_valid_day(day):
                        user_events.change_day(event_id, day)

    # Otherwise, we check for user input:
    else:
        # Wait for user to press a key:
        screen.key = stdscr.getkey()

        # If we need to select an event, change to selection mode:
        selection_keys = ['h', 'l', 'u', 'i', 'd', 'x', 'e', 'r', 'c', 'm', 'M', '.']
        if screen.key in selection_keys and user_events.filter_events_that_day(screen).items:
            screen.selection_mode = True

        # Navigation:
        if screen.key in ["n", "j", "KEY_UP", "KEY_RIGHT"]:
            screen.next_day()
        if screen.key in ["p", "k", "KEY_DOWN", "KEY_LEFT"]:
            screen.previous_day()
        if screen.key in ["KEY_HOME", "R"]:
            screen.reset_to_today()

        # Add single event:
        if screen.key == "a":
            name = input_string(stdscr, screen.y_max-2, 0, MSG_EVENT_TITLE, screen.x_max-len(MSG_EVENT_TITLE)-2)
            item_id = user_events.items[-1].item_id + 1 if not user_events.is_empty() else 1
            user_events.add_item(UserEvent(item_id, screen.year, screen.month, screen.day, name, 1, Frequency.ONCE, Status.NORMAL, False))

        # Add a recurring event:
        if screen.key == "A":
            name = input_string(stdscr, screen.y_max-2, 0, MSG_EVENT_TITLE, screen.x_max-len(MSG_EVENT_TITLE)-2)
            item_id = user_events.items[-1].item_id + 1 if not user_events.is_empty() else 1
            reps = input_integer(stdscr, screen.y_max-2, 0, MSG_EVENT_REP)
            freq = input_frequency(stdscr, screen.y_max-2, 0, MSG_EVENT_FR)
            if reps is not None and freq is not None:
                reps = 1 if reps == 0 else reps
                user_events.add_item(UserEvent(item_id, screen.year, screen.month, screen.day, name, reps+1, freq, Status.NORMAL, False))

        # Import from calcurse:
        if screen.key == "C":
            confirmed = ask_confirmation(stdscr, MSG_EVENT_IMP, cf.ASK_CONFIRMATIONS)
            if confirmed:
                importer.import_events_from_calcurse()

        # Handle "g" and "G" as go to selected (prefilled) day:
        if screen.key == "g":
            clear_line(stdscr, screen.y_max-2, 0)
            year, month, day = input_date(stdscr, screen.y_max-2, 0, MSG_GOTO)
            if screen.is_valid_date(year, month, day):
                screen.day = day
                screen.month = month
                screen.year = year
                screen.calendar_state = CalState.DAILY

        if screen.key == "G":
            clear_line(stdscr, screen.y_max-2, 0)
            question = f"{MSG_GOTO_D} {screen.year}/{screen.month}/"
            day = input_day(stdscr, screen.y_max-2, 0, question)
            if screen.is_valid_date(screen.year, screen.month, day):
                screen.day = day
                screen.calendar_state = CalState.DAILY

        # Reload:
        if screen.key == "Q":
            screen.reload_data = True
            screen.refresh_now = True

        # Change the view to monthly:
        if screen.key == "v":
            screen.calendar_state = CalState.MONTHLY

        # Other actions:
        if vim_style_exit(stdscr, screen):
            confirmed = ask_confirmation(stdscr, MSG_EXIT, cf.ASK_CONFIRMATION_TO_QUIT)
            screen.state = AppState.EXIT if confirmed else screen.state
        if screen.key == "*":
            screen.privacy = not screen.privacy
        if screen.key in [" ", "KEY_BTAB"]:
            screen.state = AppState.JOURNAL
        if screen.key == "?":
            screen.state = AppState.HELP
        if screen.key == "q":
            confirmed = ask_confirmation(stdscr, MSG_EXIT, cf.ASK_CONFIRMATION_TO_QUIT)
            screen.state = AppState.EXIT if confirmed else screen.state
        if screen.key in ["/"]:
            screen.split = not screen.split
            screen.refresh_now = True


@safe_run
def control_journal_screen(stdscr, screen, user_tasks, importer):
    """Process user input on the journal screen"""
    # If we previously selected a task, now we perform the action:
    if screen.selection_mode:

        # Timer operations:
        if screen.key == 't':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TM_ADD)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                if cf.ONE_TIMER_AT_A_TIME:
                    user_tasks.pause_all_other_timers(task_id)
                user_tasks.add_timestamp_for_task(task_id)

        if screen.key == 'T':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TM_RESET)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                user_tasks.reset_timer_for_task(task_id)

        # Add deadline:
        if screen.key == "f":
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_DEAD_ADD)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                clear_line(stdscr, screen.y_max-2, 0)
                year, month, day = input_date(stdscr, screen.y_max-2, 0, MSG_TS_DEAD_DATE)
                if screen.is_valid_date(year, month, day):
                    user_tasks.change_deadline(task_id, year, month, day)

        # Remove deadline:
        if screen.key == "F":
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_DEAD_DEL)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                user_tasks.change_deadline(task_id, 0, 0, 0)

        # Change the status:
        if screen.key in ['i', 'h']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_HIGH)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                user_tasks.toggle_item_status(task_id, Status.IMPORTANT)
        if screen.key == 'l':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_LOW)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                user_tasks.toggle_item_status(task_id, Status.UNIMPORTANT)
        if screen.key == 'u':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_RES)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                user_tasks.toggle_item_status(task_id, Status.NORMAL)
        if screen.key in ['d', 'v']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_DONE)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                user_tasks.toggle_item_status(task_id, Status.DONE)

        # Toggle task privacy:
        if screen.key == '.':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_PRIVACY)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                user_tasks.toggle_item_privacy(task_id)

        # Modify the task:
        if screen.key in ['x']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_DEL)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                user_tasks.delete_item(task_id)
        if screen.key == 'm':
            number_from = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_MOVE)
            if user_tasks.is_valid_number(number_from):
                clear_line(stdscr, screen.y_max-2)
                number_to = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_MOVE_TO)
                user_tasks.move_task(number_from, number_to)
        if screen.key in ['e', 'r']:
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_EDIT)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                clear_line(stdscr, number+2, screen.x_min)
                new_name = input_string(stdscr, number+2, screen.x_min, cf.TODO_ICON+' ', screen.x_max-4)
                user_tasks.rename_item(task_id, new_name)

        # Subtask operations:
        if screen.key == 's':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_TOG)
            if user_tasks.is_valid_number(number):
                task_id = user_tasks.items[number].item_id
                user_tasks.toggle_subtask_state(task_id)
        if screen.key == 'A':
            number = input_integer(stdscr, screen.y_max-2, 0, MSG_TS_SUB)
            if user_tasks.is_valid_number(number):
                clear_line(stdscr, screen.y_max-2, 0)
                task_name = input_string(stdscr, screen.y_max-2, 0, MSG_TS_TITLE, screen.x_max-len(MSG_TS_TITLE)-2)
                task_id = user_tasks.generate_id()
                user_tasks.add_subtask(Task(task_id, task_name, Status.NORMAL, Timer([]), False), number)
        screen.selection_mode = False

    # Otherwise, we check for user input:
    else:
        # Wait for user to press a key:
        screen.key = stdscr.getkey()

        # If we need to select a task, change to selection mode:
        selection_keys = ['t', 'T', 'h', 'l', 'v', 'u', 'i', 's', 'd', 'x', 'e', 'r', 'c', 'A', 'm', '.', 'f', 'F']
        if screen.key in selection_keys and user_tasks.items:
            screen.selection_mode = True

        # Add single task:
        if screen.key == "a":
            clear_line(stdscr, len(user_tasks.items) + 2, screen.x_min)
            task_name = input_string(stdscr, len(user_tasks.items) + 2, screen.x_min, cf.TODO_ICON+' ', screen.x_max - 4)
            task_id = user_tasks.generate_id()
            user_tasks.add_item(Task(task_id, task_name, Status.NORMAL, Timer([]), False))

        # Bulk operations:
        if screen.key in ["V", "D"]:
            confirmed = ask_confirmation(stdscr, MSG_TS_EDT_ALL, cf.ASK_CONFIRMATIONS)
            if confirmed:
                user_tasks.change_all_statuses(Status.DONE)
        if screen.key == "U":
            confirmed = ask_confirmation(stdscr, MSG_TS_EDT_ALL, cf.ASK_CONFIRMATIONS)
            if confirmed:
                user_tasks.change_all_statuses(Status.NORMAL)
        if screen.key == "L":
            confirmed = ask_confirmation(stdscr, MSG_TS_EDT_ALL, cf.ASK_CONFIRMATIONS)
            if confirmed:
                user_tasks.change_all_statuses(Status.UNIMPORTANT)
        if screen.key in ["I", "H"]:
            confirmed = ask_confirmation(stdscr, MSG_TS_EDT_ALL, cf.ASK_CONFIRMATIONS)
            if confirmed:
                user_tasks.change_all_statuses(Status.IMPORTANT)
        if screen.key in ["X"]:
            confirmed = ask_confirmation(stdscr, MSG_TS_DEL_ALL, cf.ASK_CONFIRMATIONS)
            if confirmed:
                user_tasks.delete_all_items()

        # Imports:
        if screen.key == "C":
            confirmed = ask_confirmation(stdscr, MSG_TS_IM, cf.ASK_CONFIRMATIONS)
            if confirmed:
                importer.import_tasks_from_calcurse()
                screen.refresh_now = True
        # if screen.key == "W":
            # confirmed = ask_confirmation(stdscr, MSG_TS_TW, cf.ASK_CONFIRMATIONS)
            # if confirmed:
                # importer.import_tasks_from_taskwarrior()
                # screen.refresh_now = True

        # Reload:
        if screen.key in ["Q"]:
            screen.reload_data = True
            screen.refresh_now = True

        # Other actions:
        if vim_style_exit(stdscr, screen):
            confirmed = ask_confirmation(stdscr, MSG_EXIT, cf.ASK_CONFIRMATION_TO_QUIT)
            screen.state = AppState.EXIT if confirmed else screen.state
        if screen.key == "*":
            screen.privacy = not screen.privacy
        if screen.key in [" ", "KEY_BTAB"]:
            screen.state = AppState.CALENDAR
        if screen.key == "?":
            screen.state = AppState.HELP
        if screen.key == "q":
            confirmed = ask_confirmation(stdscr, MSG_EXIT, cf.ASK_CONFIRMATION_TO_QUIT)
            screen.state = AppState.EXIT if confirmed else screen.state
        if screen.key in ["/"]:
            screen.split = not screen.split
            screen.refresh_now = True


@safe_run
def control_help_screen(stdscr, screen):
    """Process user input on the help screen"""
    # Getting user's input:
    screen.key = stdscr.getkey()

    # Handle vim-style exit on "ZZ" and "ZQ":
    if vim_style_exit(stdscr, screen):
        confirmed = ask_confirmation(stdscr, MSG_EXIT, cf.ASK_CONFIRMATION_TO_QUIT)
        screen.state = AppState.EXIT if confirmed else screen.state

    # Handle keys to exit the help screen:
    if screen.key in [" ", "?", "q", "^[", "\x7f"]:
        screen.state = AppState.CALENDAR


@safe_run
def control_welcome_screen(stdscr, screen):
    """Process user input on the welcome screen"""
    # Getting user's input:
    screen.key = stdscr.getkey()

    # Handle key to call help screen:
    if screen.key in ["?"]:
        screen.state = AppState.HELP

    # Otherwise, just start the program:
    else:
        screen.state = AppState.CALENDAR



from pathlib import Path
import re
import setuptools

setup_dir = Path(__file__).resolve().parent

version = re.search( r'__version__ = "(.*)"', Path(setup_dir, 'calcure/__main__.py').open().read())
if version is None:
    raise SystemExit("Could not determine version to use")
version = version.group(1)

setuptools.setup(
    name='calcure',
    author='Roman Anufriev',
    author_email='anufriev.roman@protonmail.com',
    url='https://github.com/anufrievroman/calcure',
    description='Modern TUI calendar and task manager',
    long_description=Path(setup_dir, 'README.md').open().read(),
    long_description_content_type='text/markdown',
    license='MIT',
    entry_points={
        "console_scripts": [
            "calcure = calcure.__main__:cli"
        ]
    },
    install_requires=['holidays', 'jdatetime', 'icalendar', 'taskw'],
    version=version,
    python_requires='~=3.7',
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Utilities",
    ],
    packages=["calcure", "calcure.translations"],
    include_package_data=True,
    data_files=[
        ('share/man/man1',
         ['calcure.1.gz']
         ),
    ],
)



"""Module that controls logging and error output"""

import logging
import io


class Error():
    """Error messages displayed to the user.
    Different errors have different purposes:
    ERROR type is reserved for loading errors;
    WARNING type is reserved for incorrect user input;"""

    def __init__(self, file):
        self.buffer = io.StringIO()
        self.file = file

        # Start logging errors:
        logging.basicConfig(level=logging.INFO,
                            format="[%(levelname)s] %(message)s",
                            # encoding='utf-8',
                            handlers=[logging.FileHandler(self.file, 'w'),
                                      #logging.StreamHandler(),
                                      logging.StreamHandler(self.buffer),
                                      ])
    @property
    def has_occured(self):
        """Has any errors occured?"""
        return self.buffer.getvalue() != ""

    @property
    def number_of_errors(self):
        """Return the number of different errors in the buffer"""
        return self.buffer.getvalue().count("[")

    @property
    def text(self):
        """Return the string with the text of the error"""
        return self.buffer.getvalue().split("] ")[1]

    @property
    def type(self):
        """Return the string with type of the error"""
        return self.buffer.getvalue().split("] ")[0]

    def clear_buffer(self):
        """Clear the buffer containing errors"""
        self.buffer.seek(0)
        self.buffer.truncate(0)

